import abc
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import numpy as np
from numpy.typing import ArrayLike, NDArray

from boss.pp.mesh import Mesh


class TestFunc(abc.ABC):
    """Base class for benchmark optimization problems."""

    # Should be truthy if the test function can accept multiple
    # data points, given as rows in an ndarray, and return a corresponding
    # array of evalutations.
    vectorized = False

    def __init__(
        self,
        std: Optional[float] = None,
        seed: Optional[int] = None,
        dim: Optional[int] = None,
        bounds: Optional[np.ndarray] = None,
    ) -> None:
        # Noisy evaluations
        self.std = std
        self.seed = seed
        if seed:
            np.random.seed(seed)

        # Default value of the bounds attribute
        if bounds is not None:
            bounds = np.asarray(bounds)

        self._bounds = bounds

        # Default value of the dim attribute
        if dim:
            self._dim = dim
        elif self.bounds is not None:
            self._dim = self.bounds.shape[0]

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def tags(self) -> dict:
        """Tags to make the function searchable / categorizable."""
        return {}

    @property
    def minima(self) -> List[Tuple[np.ndarray, float]]:
        """Optional, a list of tuples (x_min, ymin) containing any local or global minima."""
        raise NotImplementedError

    @property
    def bounds(self) -> np.ndarray:
        """The bounds of the problem, give row-wise in an ndarray."""
        return self._bounds

    @bounds.setter
    def bounds(self, new_bounds: np.ndarray) -> None:
        new_bounds = np.asarray(new_bounds)
        if self.dim is not None:
            if new_bounds.shape[0] != self.dim:
                raise ValueError(
                    f"Specified bounds {new_bounds}"
                    + f"are not consistent with dim = {self.dim}"
                )
        self._bounds = new_bounds

    @property
    @abc.abstractmethod
    def f_min(self) -> float:
        """The global minimum value."""
        raise ValueError

    @property
    @abc.abstractmethod
    def x_mins(self) -> np.ndarray:
        """The global minima given as rows in a 2d np.ndarray."""
        pass

    @property
    def x_min(self) -> np.ndarray:
        """The global minimum, if a single one exists"""
        if self.x_mins.shape[0] == 1:
            return self.x_mins[0, :]
        else:
            raise ValueError

    @property
    def f_range(self) -> np.ndarray:
        """Optional, the expected range over which f varies.

        Only used in BOSS to initialize hyperparameters."""
        pass

    def eval(self, x: np.ndarray):
        """Computation of the test function."""
        raise NotImplementedError

    def eval_grad(self, x: np.ndarray):
        """Computation of the gradient."""
        raise NotImplementedError

    def eval_with_grad(self, x: np.ndarray):
        """Simultaneous computation of function and gradient."""
        raise NotImplementedError

    @property
    def dim(self) -> int:
        """The dimension of the function domain."""
        return self._dim

    def __call__(self, x: np.ndarray) -> np.ndarray:
        if self.vectorized:
            x = np.atleast_2d(x)
        else:
            x = np.asarray(x)

        try:
            y = self.eval(x)
        except NotImplementedError:
            y, _ = self.eval_with_grad(x)

        if self.std:
            y += np.random.normal(0, self.std, size=y.shape)
        return np.squeeze(y)

    def grad(self, x: np.ndarray):
        if self.vectorized:
            x = np.atleast_2d(x)
        else:
            x = np.asarray(x)

        try:
            dy = self.eval_grad(x)
        except NotImplementedError:
            _, dy = self.eval_with_grad(x)

        if self.std:
            dy += np.random.normal(0, self.std, size=x.shape)
        return np.squeeze(dy)

    def with_grad(self, x):
        if self.vectorized:
            x = np.atleast_2d(x)
        else:
            x = np.asarray(x)

        try:
            y, dy = self.eval_with_grad(x)
        except NotImplementedError:
            y = self.eval(x)
            dy = self.eval_grad(x)

        if self.std:
            y += np.random.normal(0, self.std, size=y.shape)
            dy += np.random.normal(0, self.std, size=dy.shape)
        return np.squeeze(y), np.squeeze(dy)

    def plot(
        self,
        free_dims: ArrayLike | None = None,
        fixed_dim_vals: ArrayLike | None = None,
        grid_pts: ArrayLike = (50,),
        surface: bool = False,
        bounds: ArrayLike | None = None,
        include_min: bool = True,
        save_as: str | Path = "",
        show: bool = True,
    ):
        """
        Plots the test function.

        Parameters
        ----------
        free_dims : 1D ArrayLike of ints
            Plot using these dimensions, only relevant for >2D functions. By default
            the first two dimensions are used. Non-free dimensions are fixed to the
            corresponding components of the (first) global min.
        fixed_dim_vals : 1D ArrayLike of floats
            Coordinate values for all fixed dimensions, i.e. those
            we are not plotting over.
        grid_pts : 1D ArrayLike of ints
            Number of grid points per dimension to use when plotting.
        surface : bool
            For >=2D functions, decides whether to use a 3D surface plot
            or a contour plot.
        bounds : 2D ArrayLike of floats
            Alternative bounds, if None the test functions default bounds are used.
        save_as : str | Path
            Save the plot to this filename.
        show : bool
            If false, do not show the plot.

        Returns
        -------
        fig : Figures
            Matplotlib figure object for the plot.
        ax : Axes
            Matplotlib axes object for the plot.
        """
        import matplotlib.pyplot as plt

        bounds = bounds or self.bounds
        free_dims = free_dims or np.arange(len(bounds))[:2]

        bounds = np.atleast_2d(bounds)
        free_dims = np.asarray(free_dims)
        grid_pts = np.asarray(grid_pts)
        fixed_dim_vals = np.asarray(fixed_dim_vals)
        mesher = Mesh(bounds, free_dims, grid_pts)

        # Handle fixed dimensions: fix using global min value
        # if it exists, else the midpoint of the bounds
        if len(mesher.fixed_dims) > 0:
            if fixed_dim_vals is not None:
                mesher.fix_dims(val=fixed_dim_vals)
            elif hasattr(self, "x_mins"):
                mesher.fix_dims(self.x_mins[0, mesher.fixed_dims])
            else:
                mesher.fix_dims(
                    0.5 * np.sum(mesher.bounds, axis=1)[mesher.fixed_dims]
                )

        grid = mesher.grid
        Y = mesher.evaluate_func(self)

        # 1D plot
        if len(mesher.active_dims) == 1:
            fig, ax = plt.subplots()
            ax.plot(*grid, Y)
            ax.set(xlabel="x", ylabel="f(x)", title=self.name)
            if include_min:
                x_mins = self.x_mins[:, free_dims]
                ax.scatter(
                    x_mins,
                    self.f_min * np.ones(len(x_mins)),
                    color="red",
                    marker="o",
                    zorder=10,
                    label="Global min.",
                )
                ax.legend()
        # contour and surface plots
        elif len(mesher.active_dims) == 2:
            if surface:
                fig, ax = plt.subplots(
                    subplot_kw={"projection": "3d", "computed_zorder": False}
                )
                mappable = ax.plot_surface(
                    *grid,
                    Y,
                    cmap="viridis",
                    edgecolors="none",
                    linewidth=0,
                    zorder=1,
                )
                # plt.colorbar(mappable, ax=ax, shrink=0.7, pad=0.1)
                ax.set_zlabel(r"$f(\mathbf{x})$")
                if include_min:
                    x_mins = self.x_mins[:, free_dims]
                    ax.scatter(
                        x_mins[:, 0],
                        x_mins[:, 1],
                        self.f_min * np.ones(len(self.x_mins)),
                        s=20,
                        label="Global min.",
                        marker="o",
                        color="red",
                        zorder=10,
                    )
                    ax.legend()
            else:
                fig, ax = plt.subplots()
                mappable = ax.contourf(*grid, Y, levels=30)
                plt.colorbar(mappable, ax=ax, label=r"$f(\mathbf{x})$")
                if include_min:
                    x_mins = self.x_mins[:, free_dims]
                    ax.scatter(
                        x_mins[:, 0],
                        x_mins[:, 1],
                        s=20,
                        label="Global min.",
                        marker="o",
                        color="red",
                    )
                    ax.legend()

            ax.set(
                xlim=bounds[0],
                ylim=bounds[1],
                xlabel=r"$x_{%s}$" % free_dims[0],
                ylabel=r"$x_{%s}$" % free_dims[1],
                title=self.name,
            )
        else:
            raise ValueError("Too many free dimensions to plot")

        if save_as:
            save_as = Path(save_as)
            if not save_as.suffix:
                save_as = save_as.with_suffix(".png")
            plt.savefig(save_as, dpi=300, bbox_inches="tight", transparent=False)

        if show:
            plt.show()
        return fig, ax


def get_test_func(name: str, **kwargs) -> TestFunc:
    """Convenience instantiation of any test function."""
    test_funcs = {
        "forrester": Forrester,
        "grammacylee": GrammacyLee,
        "ackely1": Ackley1,
        "alpine1": Alpine1,
        "alpine2": Alpine2,
        "adjiman": Adjiman,
        "bartelsconn": BartelsConn,
        "beale": Beale,
        "camelthreehump": CamelThreeHump,
        "camelsixhump": CamelSixHump,
        "eggcrate": EggCrate,
        "exponential": Exponential,
        "goldsteinprice": GoldsteinPrice,
        "himmelblau": Himmelblau,
        "himmelvalley": Himmelvalley,
        "periodic": Periodic,
        "rastigrin": Rastigrin,
        "rosenbrock": Rosenbrock,
        "salomon": Salomon,
        "sphere": Sphere,
        "styblinskitang": StyblinskiTang,
        "wolfe": Wolfe,
    }
    name_lower = name.lower()
    if name_lower in test_funcs:
        f = test_funcs[name_lower](**kwargs)
    else:
        raise ValueError(f"Could not find test function {name}.")
    return f


class Forrester(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "multimodal": True,
        "dimension": 1,
    }

    bounds = np.array([[0.0, 1.0]])
    f_range = np.array([-10.0, 10.0])
    x_mins = np.array([[0.757249]])
    f_min = -6.02074
    dim = 1

    def eval(self, x):
        return (6 * x - 2) ** 2 * np.sin(12 * x - 4)

    def eval_grad(self, x):
        grad = 12 * (6 * x - 2)
        grad *= np.sin(12 * x - 4) + (6 * x - 2) * np.cos(12 * x - 4)
        return grad


class GrammacyLee(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 1,
    }

    bounds = np.array([[0.5, 2.5]])
    f_range = np.array([-10.0, 10.0])
    x_mins = np.array([[0.548563]])
    f_min = -0.869011
    dim = 1

    def eval(self, x):
        return 0.5 * np.sin(10 * np.pi * x) / x + (x - 1) ** 4


class Ackley1(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    @property
    def bounds(self):
        return np.tile([-35.0, 35.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.zeros((1, self.dim))

    @property
    def f_min(self):
        return 0.0

    def eval(self, x):
        d = self.dim
        val = (
            -20 * np.exp(-0.02 * np.sqrt(1.0 / d * np.sum(x**2, axis=1)))
            - np.exp(1.0 / d * np.sum(np.cos(2 * np.pi * x), axis=1))
            + np.exp(0)
            + 20
        )
        return val


class Alpine1(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": False,
        "dimension": None,
    }
    f_min = 0.0

    @property
    def bounds(self):
        return np.tile([-10.0, 10.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.zeros((1, self.dim))

    def eval(self, x):
        val = np.sum(np.abs(x * np.sin(x) + 0.1 * x), axis=1)
        return val


class Alpine2(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    def __init__(self, *args, **kwargs):
        if kwargs.get("dim", None) is None:
            raise ValueError(
                "dim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([0.0, 10.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.ones((1, self.dim)) * 7.917

    @property
    def f_min(self):
        return -(2.808**self.dim)

    def eval(self, x):
        val = -np.prod(np.sqrt(x) * np.sin(x), axis=1)
        return val


class Adjiman(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-1.0, 2.0], [-1.0, 1.0]])
    x_mins = np.array([[2, 0.10578]])
    f_min = -2.02181
    dim = 2

    def eval(self, x):
        val = np.cos(x[:, 0]) * np.sin(x[:, 1]) - x[:, 0] / (x[:, 1] ** 2 + 1)
        return val


class BartelsConn(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": False,
        "dimension": 2,
    }

    bounds = np.array([[-500.0, 500.0], [-500.0, 500.0]])
    x_mins = np.array([[0.0, 0.0]])
    f_min = 1.0
    dim = 2

    def eval(self, x):
        val = (
            np.abs(x[:, 0] ** 2 + x[:, 1] ** 2 + x[:, 0] * x[:, 1])
            + np.abs(np.sin(x[:, 0]))
            + np.abs(np.cos(x[:, 1]))
        )
        return val


class Beale(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-4.5, 4.5]] * 2)
    x_mins = np.array([[3.0, 0.5]])
    f_min = 0.0
    dim = 2

    def eval(self, x):
        val = (
            (1.5 - x[:, 0] + x[:, 0] * x[:, 1]) ** 2
            + (2.25 - x[:, 0] + x[:, 0] * x[:, 1] ** 2) ** 2
            + (2.625 - x[:, 0] + x[:, 0] * x[:, 1] ** 3) ** 2
        )
        return val


class CamelThreeHump(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "multimodal": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0]] * 2)
    x_mins = np.array([[0.0, 0.0]])
    f_min = 0.0
    dim = 2

    def eval(self, x):
        val = (
            2 * x[:, 0] ** 2
            - 1.05 * x[:, 0] ** 4
            + x[:, 0] ** 6 / 6
            + x[:, 0] * x[:, 1]
            + x[:, 1] ** 2
        )
        return val


class CamelSixHump(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0]] * 2)
    x_mins = np.array([[-0.0898, 0.7126], [0.0898, -0.7126]])
    f_min = -1.0316
    dim = 2

    def eval(self, x):
        val = (
            (4 - 2.1 * x[:, 0] ** 2 + x[:, 0] ** 4 / 3) * x[:, 0] ** 2
            + x[:, 0] * x[:, 1]
            + (4 * x[:, 1] ** 2 - 4) * x[:, 1] ** 2
        )
        return val


class EggCrate(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    f_range = np.array([0.0, 100.0])
    bounds = np.array([[-5.0, 5.0]] * 2)
    x_mins = np.array([[0.0, 0.0]])
    f_min = 0.0
    dim = 2

    def eval(self, x):
        val = (
            x[:, 0] ** 2
            + x[:, 1] ** 2
            + 25 * (np.sin(x[:, 0]) ** 2 + np.sin(x[:, 1]) ** 2)
        )
        return val


class Exponential(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    f_min = 1.0

    @property
    def bounds(self):
        return np.tile([-1.0, 1.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.zeros((1, self.dim))

    def eval(self, x):
        val = -np.exp(-0.5 * np.sum(x**2, axis=1))
        return val


class GoldsteinPrice(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-2.0, 2.0]] * 2)
    x_mins = np.array([[0.0, -1.0]])
    f_min = 3.0
    dim = 2

    def eval(self, x):
        x0 = x[:, 0]
        x1 = x[:, 1]
        val = (
            1
            + (x0 + x1 + 1) ** 2
            * (19 - 14 * x0 + 3 * x0**2 - 14 * x1 + 6 * x0 * x1 + 3 * x1**2)
        ) * (
            30
            + (2 * x0 - 3 * x1) ** 2
            * (18 - 32 * x0 + 12 * x0**2 + 48 * x1 - 36 * x0 * x1 + 27 * x1**2)
        )
        return val


class Himmelblau(TestFunc):
    """Himmelblau function, has 4 global mins."""

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0], [-5.0, 5.0]])
    x_mins = np.array(
        [
            [3.0, 2.0],
            [-2.805118, 3.131312],
            [-3.779310, -3.283186],
            [3.584428, -1.848126],
        ]
    )
    f_min = 0.0
    dim = 2

    def eval(self, X):
        x = X[:, 0]
        y = X[:, 1]
        z = (x**2 + y - 11) ** 2 + (x + y**2 - 7) ** 2
        return z


class Himmelvalley(TestFunc):
    """Modified Himmelblau function with one global min."""

    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 2,
    }

    bounds = np.array([[-5.0, 5.0], [-5.0, 5.0]])
    x_mins = np.array([[-4.00035, -3.55142]])
    f_min = -1.463295972168
    f_range = np.array([0.0, 10.0])
    dim = 2

    def eval(self, X):
        x = X[:, 0]
        y = X[:, 1]
        z = 0.01 * ((x**2 + y - 11) ** 2 + (x + y**2 - 7) ** 2 + 20 * (x + y))
        return z


class Periodic(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    f_range = np.array([0.0, 5])
    f_min = 0.9

    def __init__(self, *args, **kwargs):
        if kwargs.get("dim", None) is None:
            raise ValueError(
                "dim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-10.0, 10.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.zeros((1, self.dim))

    def eval(self, x):
        val = 1 + np.sum(np.sin(x) ** 2, axis=1) - 0.1 * np.exp(-np.sum(x**2, axis=1))
        return val


class Rastigrin(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    f_range = np.array([0.0, 100])
    f_min = 0.0

    def __init__(self, *args, **kwargs):
        if kwargs.get("dim", None) is None:
            raise ValueError(
                "dim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-5.12, 5.12], (self.dim, 1))

    @property
    def x_mins(self):
        return np.zeros((1, self.dim))

    def eval(self, x):
        val = 10 * self.dim + np.sum(x**2 - 10 * np.cos(2 * np.pi * x), axis=1)
        return val


class Rosenbrock(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    f_min = 0.0

    @property
    def bounds(self):
        return np.tile([-30.0, 30.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.ones((1, self.dim))

    def eval(self, x):
        val = np.sum(100 * (x[:, 1:] - x[:, :-1] ** 2) ** 2 + (x[:, :-1] - 1) ** 2)
        return val


class Salomon(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    f_min = 0.0

    def __init__(self, *args, **kwargs):
        if kwargs.get("dim", None) is None:
            raise ValueError(
                "dim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-100.0, 100.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.zeros((1, self.dim))

    def eval(self, x):
        val = (
            1
            - np.cos(2 * np.pi * np.sqrt(np.sum(x**2, axis=1)))
            + 0.1 * np.sqrt(np.sum(x**2, axis=1))
        )
        return val


class Sphere(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }

    f_min = 0.0

    @property
    def bounds(self):
        return np.tile([0.0, 10.0], (self.dim, 1))

    @property
    def x_mins(self):
        return np.zeros((1, self.dim))

    def eval(self, x):
        val = np.sum(x**2, axis=1)
        return val

    def eval_grad(self, x):
        return 2 * x


class StyblinskiTang(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": None,
    }
    f_range = np.array([-50.0, 100])

    def __init__(self, *args, **kwargs):
        if kwargs.get("dim", None) is None:
            raise ValueError(
                "dim (postive integer) must be specified for test functions with arbitrary dimension."
            )
        super().__init__(*args, **kwargs)
        if self._bounds is None:
            self._bounds = np.tile([-5.0, 5.0], (self.dim, 1))

    @property
    def f_min(self):
        return -39.16599 * self.dim

    @property
    def x_mins(self):
        return -2.903534 * np.ones((1, self.dim))

    def eval(self, x):
        val = 0.5 * np.sum(x**4 - 16 * x**2 + 5 * x, axis=1)
        return val

    def eval_grad(self, x):
        return 2 * x**3 - 16 * x + 2.5


class Wolfe(TestFunc):
    vectorized = True
    tags = {
        "continuous": True,
        "differentiable": True,
        "dimension": 3,
    }

    bounds = np.array([[0.0, 2.0]] * 3)
    x_mins = np.array([[0.0, 0.0, 0.0]])
    f_min = 0.0
    dim = 3

    def eval(self, x):
        val = (
            4.0 / 3 * (x[:, 0] ** 2 + x[:, 1] ** 2 - x[:, 0] * x[:, 1]) ** 0.75
            + x[:, 2]
        )
        return val



from __future__ import annotations

import copy
from functools import reduce
from typing import Any

import GPy
import numpy as np
from GPy.kern import Kern
from numpy.typing import ArrayLike, NDArray

from boss.bo.models.model import Model
from boss.utils.arrays import shape_consistent, shape_consistent_XY


class SingleTaskModel(Model):
    """
    Functionality for creating, refitting and optimizing a GP model
    """

    def __init__(
        self,
        kernel: Kern,
        X: ArrayLike,
        Y: ArrayLike,
        noise: float = 1e-12,
        ynorm: bool = False,
        noise_optim=False,
    ) -> None:
        """
        Initializes the SingleTaskModel class.
        """
        # normalise observation mean:
        self.normmean = np.mean(Y)
        # scale normalisation is not used unless ynorm is true:
        self.use_norm = ynorm
        # previous boss code used normsd to normalise observation variance:
        # if self.ynorm: self.normsd = np.std(Y)
        # current version normalises observation range:
        self.normsd = np.ptp(Y) if self.use_norm else 1
        # note that the choice betweeen variance or range normalisation needs
        # to be taken into account when we set kernel parameter priors
        # normalised data:
        Y_norm = (Y - self.normmean) / self.normsd
        # initialise model
        self.model = GPy.models.GPRegression(X, Y_norm, kernel=kernel, noise_var=noise)

        self.noise_optim = noise_optim
        if not noise_optim:
            self.model.likelihood.fix()

    @property
    def dim(self) -> int:
        return self.model.kern.input_dim

    @property
    def kernel(self) -> Kern:
        return self.model.kern

    @property
    def X(self) -> NDArray:
        return self.model.X

    @property
    def Y(self) -> NDArray:
        return self.model.Y * self.normsd + self.normmean

    def __deepcopy__(self, memo: dict) -> SingleTaskModel:
        cls = self.__class__
        model_copy = cls.__new__(cls)
        memo[id(self)] = model_copy
        for key, val in self.__dict__.items():
            # A GPy kernel object attached to a model can't be deepcopied in the
            # usual way due to a bug so we have to use the kernel's custom copy method.
            if key == "_kernel":
                setattr(model_copy, key, val.copy())
            else:
                setattr(model_copy, key, copy.deepcopy(val, memo))
        return model_copy

    def add_data(self, X_new: ArrayLike, Y_new: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset appending.
        """
        X_new, Y_new = shape_consistent_XY(X_new, Y_new, self.dim)
        X = np.vstack([self.X, X_new])
        Y = np.vstack([self.Y, Y_new])
        self.redefine_data(X, Y)

    def redefine_data(self, X: ArrayLike, Y: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        X, Y = shape_consistent_XY(X, Y, self.dim)
        # update normalisation
        self.normmean = np.mean(Y)
        if self.use_norm:
            self.normsd = np.ptp(Y)
        # update model
        Y_norm = (Y - self.normmean) / self.normsd
        self.model.set_XY(np.atleast_2d(X), np.atleast_2d(Y_norm))

    def get_best_xy(self) -> tuple[NDArray, float]:
        """
        Returns the lowest energy acquisition (x, y).
        """
        x_best = np.array(self.X[np.argmin(self.Y)])
        y_best = np.min(self.Y)
        return x_best, y_best

    def predict(
        self, X: ArrayLike, noise: bool = True, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise).
        """
        X = shape_consistent(X, self.dim)
        m, v = self.model.predict(X, include_likelihood=noise)
        v = np.clip(v, 1e-12, np.inf)
        if norm:
            return m, v
        return m * self.normsd + self.normmean, v * (self.normsd**2)

    def predict_grads(
        self, X: ArrayLike, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance gradients with respect to
        input at point x.
        """
        X = shape_consistent(X, self.dim)
        dmdx, dvdx = self.model.predictive_gradients(X)
        if norm:
            return dmdx, dvdx
        return dmdx * self.normsd, dvdx * (self.normsd**2)

    def predict_mean_sd_grads(
        self, X: ArrayLike, noise: bool = True, norm: bool = True
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise).

        This method is a wrapper used primarily during calculations
        of acquisition functions and their derivatives.
        """
        m, v = self.predict(X, noise=noise, norm=norm)
        dmdx, dvdx = self.predict_grads(X, norm=norm)
        dmdx = dmdx[:, :, 0]
        dsdx = dvdx / (2 * np.sqrt(v))
        return m, np.sqrt(v), dmdx, dsdx

    def predict_mean_grad(
        self, X: ArrayLike, norm: bool = True
    ) -> tuple[NDArray, NDArray]:
        """Returns model mean and its gradient at point x.

        This method is a wrapper used primarily when the mean function
        is minimized in order to obtain a global minimum prediction.
        """
        m, _ = self.predict(X, norm=norm)
        dmdx, _ = self.predict_grads(X, norm=norm)
        return m, dmdx

    def estimate_num_local_minima(self, search_bounds: ArrayLike) -> int:
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        # For the ith dimension, the number of local minima along a slice
        # is approximately n(i) = boundlength(i)/(2*lengthscale(i)). Note
        # that periodic kernels operate on normalised distances: distance
        # between inputs that are period(i)/2 apart is 1. To get the total
        # number of minima for all of the search space, multiply together
        # n(i) over all i.
        search_bounds = np.atleast_2d(search_bounds)
        numpts = 1
        ks = self.model.kern.parameters if self.dim > 1 else [self.model.kern]
        for bounds, kern in zip(search_bounds, ks):
            if hasattr(kern, "period"):
                bound_distance = (bounds[1] - bounds[0]) / kern.period[0]
            else:
                bound_distance = (bounds[1] - bounds[0]) / 2
            numpts *= max(1, bound_distance / kern.lengthscale[0])
        return int(numpts)

    @staticmethod
    def _rename_param(name: str) -> str:
        substitutions = {
            "kern": "kernel",
            "het_Gauss": "noise",
            "Gaussian_noise": "noise",
        }
        new_name = name
        for text, sub in substitutions.items():
            new_name = new_name.replace(text, sub)

        return new_name

    def get_all_params(self, include_fixed: bool = True) -> dict[str, Any]:
        """
        Get all model parameter names and values.

        Parameter names returned this way are changed to conform
        with BOSS' nomenclature and do not in general correspond
        to the model-internal parameter names.

        Parameters
        ----------
        include_fixed : bool
            Whether to include fixed parameters, defaults to True.

        Returns
        -------
        dict[str, NDArray]
            Dict mapping parameter names to their corresponding values.
        """
        model = self.model
        param_names = model.parameter_names()
        params = {}
        is_unfixed = lambda par: "fixed" not in par.constraints.properties()
        for name in param_names:
            # separate kernel_name from para_name and get the
            # actual parameter object by repeatedly applying getattr
            # this is 100x faster than model[name] which uses a regex search
            obj_names = name.split(".")
            par = reduce(lambda x, y: getattr(x, y), obj_names, model)
            if is_unfixed(par) or include_fixed:
                new_name = ".".join(obj_names[-2:])
                new_name = self._rename_param(new_name)
                params[new_name] = np.squeeze(par.values)

        return params

    def get_unfixed_params(self) -> NDArray:
        """
        Returns the unfixed parameters of the model in an array.
        """
        return np.array(self.model.unfixed_param_array.copy()).astype(float)

    def sample_unfixed_params(self, num_samples: int) -> NDArray:
        """
        Sample unfixed model parameters.
        """
        hmc = GPy.inference.mcmc.HMC(self.model)
        burnin = hmc.sample(int(num_samples * 0.33))
        return hmc.sample(num_samples)

    def set_unfixed_params(self, params: NDArray) -> None:
        """
        Sets the unfixed parameters of the model to given values.
        """
        self.model[self.model._fixes_] = params
        self.model.parameters_changed()

    @property
    def noise(self):
        return self.model.likelihood[0]

    @property
    def lengthscales(self) -> NDArray:
        params = self.get_all_params(include_fixed=True)
        values = [val for name, val in params.items() if "lengthscale" in name]
        return np.array(values)

    def optimize(self, restarts: int = 1) -> None:
        """
        Updates the model hyperparameters by maximizing marginal likelihood.
        """
        self.model.optimization_runs = []
        if restarts == 1:
            self.model.optimize()
        else:
            self.model.optimize_restarts(
                num_restarts=restarts, verbose=False, messages=False
            )



import os
import platform
from enum import Enum
from pathlib import Path
from typing import Any, Literal

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import rcParams
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.typing import ArrayLike, NDArray


class Tableau20(str, Enum):
    blue = "#1F77B4"
    red = "#D62728"
    green = "#2CA02C"
    orange = "#FF7F0E"
    purple = "#9467BD"
    brown = "#8C564B"
    pink = "#E377C2"
    grey = "#7F7F7F"
    yellow = "#BCBD22"
    turquoise = "#17BECF"
    lightblue = "#AEC7E8"
    lightred = "#FF9896"
    lightorange = "#FFBB78"
    lightgreen = "#98DF8A"
    lightpurple = "#C5B0D5"
    lightbrown = "#C49C94"
    lightpink = "#F7B6D2"
    lightgrey = "#C7C7C7"
    lightyellow = "#DBDB8D"
    lightturquoise = "#9EDAE5"


DEFAULT_STYLE = {
    # Fonts and sizes
    "font.size": 14,
    "axes.titlesize": 16,
    "axes.titlepad": 10,
    "axes.labelsize": 14,
    "legend.fontsize": 12,
    "xtick.labelsize": 12,
    "ytick.labelsize": 12,
    # Axes properties
    "axes.linewidth": 1.5,
    "axes.edgecolor": "black",
    # Ticks
    "xtick.direction": "in",
    "ytick.direction": "in",
    "xtick.major.width": 1.2,
    "ytick.major.width": 1.2,
    "xtick.minor.width": 0.9,
    "ytick.minor.width": 0.9,
    "xtick.major.size": 6,
    "ytick.major.size": 6,
    "xtick.minor.size": 4,
    "ytick.minor.size": 4,
    "xtick.major.pad": 8,
    "ytick.major.pad": 8,
    # Enabling ticks on all sides
    "axes.spines.top": True,
    "axes.spines.right": True,
    "xtick.top": True,
    "ytick.right": True,
    "xtick.minor.top": True,
    "ytick.minor.right": True,
    # Subplot adjustments
    # "figure.subplot.left": 0.14,       # Default left margin
    # "figure.subplot.right": 0.95,    # Default right margin
    # "figure.subplot.bottom": 0.14,    # Default bottom margin
    # "figure.subplot.top": 0.9,       # Default top margin
    # "figure.subplot.wspace": 0.2,    # Default width space
    # "figure.subplot.hspace": 0.2,    # Default height space
    # "figure.constrained_layout.use": True,
    "figure.autolayout": True,
    # Grid
    # "axes.grid": True,               # Enable grid
    # "axes.facecolor": "#f0f0f0",     # Light grey background
    # "axes.grid.axis": "both",        # Grid for both axes
    # "axes.grid.which": "major",      # Major grid lines
    # "grid.color": "white",           # Gridline color
    # "grid.linestyle": "-",           # Solid gridlines
    # "grid.linewidth": 1,             # Gridline width
    # Lines and markers
    "lines.linewidth": 2,  # Line width
    "lines.markersize": 8,  # Marker size
    # Figure properties
    "savefig.dpi": 300,
    "savefig.bbox": "tight",
    "figure.figsize": (6, 4),
    # Disable frame on legend
    "legend.frameon": True,
    # Default color cycle
    # "axes.prop_cycle": rcParams['axes.prop_cycle'],
    # Additional spacing
    "axes.labelpad": 6,
}


def get_config_dir() -> Path:
    """
    Returns the default directory for configuration files based on the operating system.

    For Linux: ~/.config/<application_name>
    For macOS: ~/Library/Application Support/<application_name>
    For Windows: %APPDATA%\\<application_name>

    Returns
    -------
    config_dir: Path
        The default configuration directory for the operating system.
    """
    system = platform.system()

    if system == "Linux":
        config_dir = Path.home() / ".config"
    elif system == "Darwin":  # macOS
        config_dir = Path.home() / "Library" / "Application Support"
    elif system == "Windows":
        config_dir = Path(
            os.environ.get("APPDATA", Path.home() / "AppData" / "Roaming")
        )
    else:
        raise NotImplementedError(f"Unsupported operating system: {system}")

    return config_dir


def load_style() -> None:
    """
    Loads a style sheet for matplotlib figures from a user-created
    config file SYSTEM_CONFIG_DIR/boss/boss.mplstyle if it exists, otherwise
    fallsback to a default style.
    """
    config_dir = get_config_dir()
    style_path = config_dir / "boss" / "boss.mplstyle"
    if style_path.exists():
        style = style_path
    else:
        style = DEFAULT_STYLE
    plt.style.use(style)



import numpy as np
from scipy.optimize import bisect, minimize
from scipy.stats import gamma, norm


def gammaparams(q1, q2, p1=0.1, p2=0.9):
    """A function for parametrizing gamma distributions by specifying two
    quantiles.

    Relevant math can be found in "Determining distribution parameters from
    quantiles" by John D. Cook, 2010.

    :param q1: p1-quantile
    :param q2: p2-quantile
    :param p1: lower percentage (0 < p1 < 1)
    :param p2: higher percentage (0 < p2 < 1)
    """
    # Require the arguments to be positive and q1 < q2, p1 < p2 <1
    if q1 <= 0 or q2 <= q1:
        return (None, None)
    if p1 <= 0 or p2 <= p1 or p2 >= 1:
        return (None, None)

    def f(a):
        return np.abs(gamma.ppf(p2, a, scale=1) / gamma.ppf(p1, a, scale=1) - q2 / q1)

    shape = minimize(f, 1, bounds=[[0.1, None]], method="L-BFGS-B").x
    rate = gamma.ppf(p1, shape, scale=1) / q1

    return (shape[0], rate[0])


def fit_gumbel(mean, sd):
    """Helper function to obtain parameters for a Gumbel distribution.
    This distribution is used to model the distribution of the
    unknown global minimum.

    The parameters are determined using mean and standard deviation values at
    random locations in the user function input space. With this, an empirical
    CDF P(Y_* < Y) can be constructed (probability of a random variable Y
    being larger than the (unknown) minimum Y_* of the user function,
    given what we know about the user function through the posterior model)

    Mathematically, if Y_* is our (unknown) global minimum value, this
    empirical CDF can be constructed via
        P(Y_* < Y) = 1 - \prod_i^N CDF(-(Y-\mu_i)/\sigma_i)
    if \mu_i and \sigma_i are the mean and standard deviations at N random
    locations from the user function input space.
    Rewrite this to stabilize the numeric computation:
        P(Y_* < Y) = 1 - exp^{ln{\sum_i^N CDF(-(Y-\mu_i)/\sigma_i)}}

    We then solve this CDF for the 25th, 50th and 75th percentile and use
    the found Y to obtain Gumbel scaling parameters.

    Parameters
    ----------
    mean : ndarray
        Mean values at num_random_samples values in the input domain
    sd : ndarray
        Corresponding standard deviations at the mean values
    """

    def empirical_cdf(Y, mean, sd):
        """Construct the empirical CDF P(Y_* < Y) using posterior statistics.

        Parameters
        ----------
        Y : numpy.float64
            Random variable Y
        mean : ndarray
            Posterior mean at num_random_samples samples
        sd : ndarray
            Posterior standard deviation at num_random_samples samples

        Returns
        -------
        numpy.float64
            Probability 0 <= p <= 1 of Y being global minimum
        """
        return 1 - np.exp(np.sum(norm.logcdf(-(Y - mean) / sd), axis=0))

    # Search interval bounds
    left, right = np.min(mean - 5 * sd), np.max(mean + 5 * sd)

    def binary_search(prob_val):
        """Helper function that applies root search to find Y where the
        probability of Y being the global minimum is prob_val.

        Parameters
        ----------
        prob_val : float
            Probability of Y being global minimum (0 <= p <= 1)

        Returns
        -------
        float
            Y value corresponding to the prob-val's quantile
        """
        return bisect(
            lambda x: empirical_cdf(x, mean, sd) - prob_val, left, right, maxiter=10000
        )

    # Search lower, middle and upper quantiles
    lower, middle, upper = [binary_search(p) for p in [0.25, 0.5, 0.75]]

    # Solve for Gumbel scaling parameters
    b = (lower - upper) / (np.log(np.log(4.0 / 3.0)) - np.log(np.log(4.0)))
    a = middle - b * np.log(np.log(2.0))
    return a, b



"""A collection of array-related functions.
"""

from __future__ import annotations

import itertools
from itertools import accumulate, chain

import numpy as np
from numpy.typing import ArrayLike, NDArray


def cartesian_product(input_arrays: tuple[NDArray, ...] | list[NDArray]) -> NDArray:
    """
    Optimized computation of cartesian products.

    Reference: https://stackoverflow.com/a/49445693
    """
    num_arrays = len(input_arrays)
    array_lengths = *map(len, input_arrays), num_arrays
    result_dtype = np.result_type(*input_arrays)
    temp_array = np.empty(array_lengths, dtype=result_dtype)

    temp_slices = (
        *accumulate(
            chain((temp_array,), itertools.repeat(0, num_arrays - 1)),
            np.ndarray.__getitem__,
        ),
    )
    indexing_tuple = slice(None), *itertools.repeat(None, num_arrays - 1)

    for array_idx in range(num_arrays - 1, 0, -1):
        temp_slices[array_idx][..., array_idx] = input_arrays[array_idx][
            indexing_tuple[: num_arrays - array_idx]
        ]
        temp_slices[array_idx - 1][1:] = temp_slices[array_idx]

    temp_array[..., 0] = input_arrays[0][indexing_tuple]
    return temp_array.reshape(-1, num_arrays)


def concatenated_cartesian_product(x: NDArray, y: NDArray) -> NDArray:
    """
    Computes the row-wise cartesian product between x and y,
    concatenating each pair in the in product.
    """
    x_repeated = np.repeat(x, repeats=y.shape[0], axis=0)
    y_tiled = np.tile(y, (x.shape[0], 1))
    return np.hstack((x_repeated, y_tiled))


def shape_consistent(arr: ArrayLike, dim: int) -> NDArray[np.float64]:
    """Ensures that an array of row vectors is consistent with a given dimension.

    Parameters
    ----------
    X : ArrayLike of floats
        Locations for user function observations.
    dim : int
        Dimension of the user function domain.

    Returns
    -------
    NDArray[np.float64]
        X-data with consistent shape.
    """
    arr = np.atleast_2d(arr)
    if dim == 1:
        if arr.shape[1] > dim:
            arr = arr.T
    else:
        if arr.shape[1] != dim:
            raise ValueError(
                f"X-shape = {arr.shape} inconsistent with dimension = {dim}"
            )
    return arr


def shape_consistent_XY(
    X: ArrayLike,
    Y: ArrayLike | None,
    x_dim: int,
    y_dim: int = 1,
    nan_pad: bool = False,
) -> tuple[NDArray, NDArray]:
    """Ensures that X and Y-data are shape consistent.

    The shape of X is checked for consistency with the user function dimension.
    The shape of Y is then checked for consistency with X and possibly padded with nan-values upon request.

    This function should be called by any user-facing method that accepts X, Y-data.

    Parameters
    ----------
    X : ArrayLike of floats
        Input data, scalars and 1d-arrays will be promoted to 2d.
    Y : ArrayLike of floats | None
        Output data, scalars and 1d-arrays will be promoted to 2d.
    x_dim : int
        Dimension of the user function domain, as specified by the bounds.
    nan_pad : bool = False
        Whether to allow Y with less rows than X to be nan-padded until
        the number of rows match.
    y_dim : int = int
        Dimension of the objective output.

    Returns
    -------
    tuple[NDArray, NDArray]:
        X, Y-data with consistent shapes.
    """
    X = shape_consistent(X, x_dim)
    if Y is None:
        if nan_pad:
            Y = np.empty((X.shape[0], y_dim)) * np.nan
        else:
            raise ValueError("Y=None not allowed for nan_pad=False")
    else:
        Y = shape_consistent(Y, y_dim)
        n_diff = X.shape[0] - Y.shape[0]
        if n_diff > 0:
            if nan_pad:
                Y_fill = np.empty((n_diff, y_dim)) * np.nan
                Y = np.concatenate((Y, Y_fill), axis=0)
            else:
                raise ValueError("Number of rows in X and Y must match.")
        elif n_diff < 0:
            raise ValueError("Y cannot contain more rows than X.")
    return X, Y



from __future__ import annotations

import copy
import os
import warnings
from collections import UserDict
from pathlib import Path
from typing import Any, Callable

import numpy as np

import boss.io.dump as dump
import boss.io.parse as parse
import boss.keywords as bkw
from boss.bo.acq.cost import CostFunc
from boss.bo.userfunc import MultiTask, SingleTask, SingleTaskGradient


class Settings(UserDict):
    """Reads, interprets and defines the code internal settings based on input."""

    def __init__(self, keywords: dict[str, Any], f: Callable | None = None) -> None:
        super().__init__()

        # Non-keyword attributes.
        self.is_rst = False

        self.dir = os.getcwd()
        self.costfn = None
        self.user_keys = list(keywords.keys())

        # Before any other keywords are set, we assign values to all keywords
        # with independent defaults.
        self.set_independent_defaults(only_missing=False)

        # Update with BOSS keywords passed to init
        self.update(keywords)
        if keywords.get("bounds") is None:
            raise ValueError("Keyword 'bounds' has to be defined by the user")

        self.deprecation_notice()
        self.correct()

        # Handle the user function: if a function is passed directly we take that,
        # otherwise fall back to the function specified by the userfn keyword.
        self.f = None

        # 1) one user function
        if f is None:
            if self["userfn"] is not None:
                f = bkw.func_from_keyword(self["userfn"])
                if self.is_grad:
                    self.f = SingleTaskGradient(f, self.dim)
                else:
                    self.f = SingleTask(f, self.dim)
        else:
            if callable(f):
                self["userfn"] = bkw.func_to_keyword(f)
                if self.is_grad:
                    self.f = SingleTaskGradient(f, self.dim)
                else:
                    self.f = SingleTask(f, self.dim)

        # 2) multiple user functions
        if f is None:
            if self["userfn_list"] is not None:
                f = [bkw.func_from_keyword(fn) for fn in self["userfn_list"]]
                self.f = MultiTask(f, self.dim)
        else:
            if isinstance(f, list):
                self["userfn_list"] = [bkw.func_to_keyword(fn) for fn in f]
                self.f = MultiTask(f, self.dim)

        # Set default values for dependent keywords if they
        # have not yet been specified.
        self.set_dependent_defaults(only_missing=True)

        # Handle cost function: take either directly passed function or
        # function specified by costfn keyword
        if isinstance(self["costfn"], str):
            costfn = bkw.func_from_keyword(self["costfn"])
            self.costfn = CostFunc(costfn, self.dim + int(self.is_multi))
        else:
            if callable(self["costfn"]):
                self.costfn = CostFunc(self["costfn"], self.dim + int(self.is_multi))
                self["costfn"] = bkw.func_to_keyword(self.costfn.func)

        # Handle heteroscedastic noise function: take directly passed function or
        # function specified by hsc_noise keyword
        if isinstance(self["hsc_noise"], str):
            self.hsc_noise = bkw.func_from_keyword(self["hsc_noise"])
        else:
            if callable(self["hsc_noise"]):
                self.hsc_noise = self["hsc_noise"]
                self["hsc_noise"] = bkw.func_to_keyword(self.hsc_noise)

        # Set RNG seed if specified.
        # TODO: Propagate this seed to GPy to eliminate all randomness.
        if self["seed"] is not None:
            np.random.seed(self["seed"])

        self.check()

    @classmethod
    def from_file(cls, file_path: str | Path) -> Settings:
        """Factory method for Constructing a Settings object from a boss input file.

        Parameters
        ----------
        file_path: str, Path
            Path to the input file.

        Returns
        -------
        Settings
            Settings object generated using the input file.
        """
        input_data = parse.parse_input_file(file_path, skip="results")
        settings = cls(input_data["keywords"])
        settings.is_rst = input_data["is_rst"]
        return settings

    def copy(self) -> Settings:
        return copy.deepcopy(self)

    def set_independent_defaults(self, only_missing: bool = True) -> None:
        """Sets default values for independent keywords."""
        if not only_missing:
            for cat in bkw.get_copied_categories():
                self.update(cat)
        else:
            for cat in bkw.get_copied_categories():
                for key, val in cat.items():
                    if self.get(key) is None:
                        self[key] = val

    def set_dependent_defaults(self, only_missing: bool = True) -> None:
        """Sets default values for keywords that depend on other keywords."""
        should_update = lambda key: self.get(key) is None or only_missing is False

        if should_update("periods"):
            self["periods"] = self["bounds"][:, 1] - self["bounds"][:, 0]
        if should_update("iterpts"):
            self["iterpts"] = int(15 * self.dim**1.5)
        if should_update("min_dist_acqs"):
            self["min_dist_acqs"] = 0.01 * min(self["periods"])

        # Default multi-task optimisation settings
        if should_update("num_tasks"):
            if self["userfn_list"] is None:
                self["num_tasks"] = 1
            else:
                self["num_tasks"] = len(self["userfn_list"])
        if should_update("task_initpts"):
            self["task_initpts"] = np.tile(self["initpts"], self["num_tasks"])
        if should_update("task_cost"):
            self["task_cost"] = np.tile(1, self["num_tasks"])

        # Scale normalisation default depends on optimisation type
        if should_update("ynorm"):
            self["ynorm"] = self.is_multi

        if self.is_multi and should_update("W_init"):
            self["W_init"] = np.zeros(self["num_tasks"] * self["W_rank"])

        # Model slice and number of points.
        if should_update("pp_model_slice"):
            if self.dim == 1:
                self["pp_model_slice"] = np.array([1, 1, 100])
            elif self.dim == 2:
                self["pp_model_slice"] = np.array([1, 2, 50])
            else:
                self["pp_model_slice"] = np.array([1, 2, 25])

    @property
    def dim(self) -> int:
        """The dimensionality of the user-supplied objective.

        The number of dimensions is a read-only propery that is
        derived from the bounds provided by the user.

        Returns
        -------
        int
            The dimensionality of the objective.

        """
        return len(self["bounds"])

    @property
    def is_grad(self) -> bool:
        return self["model_name"] == "GradientModel"

    @property
    def is_multi(self) -> bool:
        """Indicates whether multi-task data is used.

        Returns
        -------
        bool

        """
        return self["model_name"] == 'MultiTaskModel'

    @property
    def is_hsc(self) -> bool:
        """Indicates whether a heteroscedastic model is used.

        Returns
        -------
        bool

        """
        return self["model_name"] == 'HeteroscedasticModel'

    def dump(self, file_path: str | Path, only_user_keywords: bool = True) -> None:
        """Writes the current settings to a boss input file.

        Parameters
        ----------
        fname : Union[str, path]
            Path to the destination file.
        """
        if only_user_keywords:
            keywords = {k: v for k, v in self.items() if k in self.user_keys}
        else:
            keywords = self
        dump.dump_input_file(file_path, keywords)

    def correct(self) -> None:
        """Corrects the type and value of certain keywords.

        The user is afforded some laziness defining certain keywords,
        e.g., by providing lists instead of np.arrays.
        """
        # Make sure int and float arrays are np and not python sequences.
        for key, val in self.items():
            cat = bkw.find_category(key)
            cat_type, cat_dim = cat[0], cat[1]
            if cat_dim > 0 and cat_type in [int, float] and val is not None:
                self[key] = np.asarray(val, dtype=cat_type)

        model_aliases = {
            'gpy': {
                ('single_task', 'st'): 'SingleTaskModel',
                ('multi_task', 'mt'): 'MultiTaskModel',
                ('grad', 'gradient'): 'GradientModel',
                ('hsc', 'heteroscedastic'): 'HeteroscedasticModel',
            },
            'torch': {
                ('single_task', 'st'): 'SingleTaskModel',
            }
        }
        for aliases, model_class in model_aliases[self['model_backend']].items():
            if self['model_name'] == model_class:
                break
            if self['model_name'].lower() in aliases:
                self['model_name'] = model_class

        self["bounds"] = np.atleast_2d(self["bounds"])

        kernel = self["kernel"]
        if isinstance(kernel, str):
            self["kernel"] = [kernel]

        if len(self["kernel"]) == 1:
            self["kernel"] *= self.dim

    def check(self) -> None:
        if self.is_multi:
            self.check_multitask_settings()

        if self['userfn_list'] is not None and self['model_name'] != 'MultiTaskModel':
            raise ValueError('To use a multi-task model you must set model_name=multi_task')

        if self.is_hsc and not callable(self.hsc_noise):
            raise ValueError(
                "A valid heteroscedastic noise function must be provided when using a heteroscedastic model."
            )

    def check_multitask_settings(self) -> None:
        """Finds inconsistencies and possible errors in multi-task settings."""
        if self.f is not None:
            if not isinstance(self.f, MultiTask):
                raise ValueError(
                    "Multi-task optimisation requires a user function list. "
                    "Use keyword 'userfn_list' or input option f with list input."
                )

            if len(self.f) > self["num_tasks"]:
                raise ValueError(
                    "The number of user functions ({}) ".format(len(self.f))
                    + "exceeds the number of tasks ({}).".format(self["num_tasks"])
                )

            if len(self.f) < self["num_tasks"]:
                warnings.warn(
                    "The number of tasks ({}) ".format(self["num_tasks"])
                    + "exceeds the number of user functions ({}).".format(len(self.f))
                )

        if len(self["task_initpts"]) != self["num_tasks"]:
            raise ValueError(
                "The number of settings under keyword 'task_initpts' ({}) ".format(
                    len(self["task_initpts"])
                )
                + "does not match the number of tasks ({}).".format(self["num_tasks"])
            )

        if self["task_cost"] is not None:
            if len(self["task_cost"]) != self["num_tasks"]:
                raise ValueError(
                    "The number of acquisition costs ({}) ".format(
                        len(self["task_cost"])
                    )
                    + "does not match the number of tasks ({}).".format(
                        self["num_tasks"]
                    )
                )

            if np.any(self["task_cost"][1:] > self["task_cost"][0]):
                warnings.warn(
                    "The acquisition cost associated with one or more "
                    "support tasks exceeds the target acquisition cost."
                )

        if self["maxcost"] is not None:
            if self["maxcost"] < np.sum(self["task_initpts"] * self["task_cost"]):
                raise ValueError(
                    "The maximum cost is too low to accommodate for the "
                    "requested initialization data."
                )

    def deprecation_notice(self) -> None:
        deprecated = {
            "pp_truef_at_xhats": "pp_truef_at_glmins",
            "verbosity": None,
            "glmin_tol": "conv_tol and conv_patience",
            "ygrad": "model_name = 'grad'"
        }
        msg = ""
        for key, val in deprecated.items():
            if key in self:
                if val is None:
                    msg += f"Keyword {key} has been deprecated and should be removed\n"
                else:
                    msg += f"Keyword {key} has been renamed, use {val} instead\n"

        if len(msg) > 0:
            raise RuntimeError(msg)



from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any

import numpy as np
from numpy.typing import ArrayLike, NDArray

if TYPE_CHECKING:
    from boss.pp.graphics import MeshGraphic

from boss.utils.sparselist import wrap_index


class Element(abc.ABC):
    """Abstract base class for all plot elements."""

    @property
    def graphic(self) -> MeshGraphic:
        """The graphic that the element is plotted on."""
        return self._graphic

    @graphic.setter
    def graphic(self, graphic: MeshGraphic) -> None:
        self._graphic = graphic

    @abc.abstractmethod
    def plot(self):
        """Plots the element on the graphic's axes."""
        pass


class Scatter(Element):
    """An element that creates a scatter plot."""

    def __init__(
        self,
        **scatter_kwargs: Any,
    ) -> None:
        """Initialize the Scatter element.

        Parameters
        ----------
        **scatter_kwargs : Any
            Keyword arguments passed to `matplotlib.axes.Axes.scatter`.
        """
        self.scatter_kwargs = scatter_kwargs

    @property
    @abc.abstractmethod
    def scatter_data(self) -> NDArray:
        """The data to be scattered."""
        pass

    def plot(self) -> None:
        """Creates the scatter plot on the graphic's axes."""
        _, ax = self.graphic.fig, self.graphic.ax
        kwargs = {'clip_on': False, 'zorder': 10}
        kwargs.update(self.scatter_kwargs)
        ax.scatter(*self.scatter_data.T, **kwargs)


class Acquisitions(Scatter):
    """An element for plotting acquired data points."""

    def __init__(
        self,
        itr: int | NDArray | None = None,
        **scatter_kwargs: Any,
    ) -> None:
        """Initialize the Acquisitions element.

        Parameters
        ----------
        itr : int | NDArray | None, optional
            The iteration(s) to plot. If None, all acquisitions up to the
            current iteration are plotted.
        **scatter_kwargs : Any
            Keyword arguments passed to `matplotlib.axes.Axes.scatter`.
        """
        self.itr = itr
        super().__init__(**scatter_kwargs)

    @property
    def scatter_data(self) -> NDArray:
        """The acquisition data to be plotted."""
        graphic = self.graphic
        # If an iteration was not specified we plot all acquisitions
        # from start to current iteration.
        if self.itr is None:
            itr_curr = wrap_index(graphic.itr, graphic.results.num_iters)
            self.itr = np.arange(itr_curr + 1) if itr_curr > 0 else 0

        results = graphic.results
        X = results.select("X", itr=self.itr)[:, graphic.active_input_dims]
        Y = results.select("Y", itr=self.itr)[:, graphic.active_target_dims]
        acq = np.hstack((X, Y))[:, : graphic.NUM_AXIS]
        return acq


class NextAcquisition(Scatter):
    """An element for plotting the next proposed acquisition point."""

    def __init__(
        self,
        **scatter_kwargs: Any,
    ) -> None:
        """Initialize the NextAcquisition element.

        Parameters
        ----------
        **scatter_kwargs : Any
            Keyword arguments passed to `matplotlib.axes.Axes.scatter`.
        """
        super().__init__(**scatter_kwargs)

    @property
    def scatter_data(self) -> NDArray:
        """The next acquisition point to be plotted."""
        graphic = self.graphic
        res = graphic.results
        X = res.get_next_acq(self.graphic.itr)[:, graphic.active_input_dims]
        return X


class Minimum(Scatter):
    """An element for plotting the inferred global minimum."""

    def __init__(
        self,
        itr: int | NDArray | None = None,
        **scatter_kwargs: Any,
    ) -> None:
        """Initialize the Minimum element.

        Parameters
        ----------
        itr : int | NDArray | None, optional
            The iteration to plot the minimum for. If None, the current
            iteration is used.
        **scatter_kwargs : Any
            Keyword arguments passed to `matplotlib.axes.Axes.scatter`.
        """
        self.itr = itr
        super().__init__(**scatter_kwargs)

    @property
    def scatter_data(self) -> NDArray:
        """The inferred global minimum to be plotted."""
        graphic = self.graphic
        results = graphic.results
        self.itr = self.itr or graphic.itr
        x_glmin = np.atleast_2d(results.select("x_glmin", itr=self.itr))
        mu_glmin = np.atleast_2d(results.select("mu_glmin", itr=self.itr)).T
        minimum = np.hstack((x_glmin[:, graphic.active_input_dims], mu_glmin))
        return minimum[:, : graphic.NUM_AXIS]


class BestAcquisition(Scatter):
    """An element for plotting the best acquisition so far."""

    def __init__(
        self,
        itr_max: int | None = None,
        **scatter_kwargs: Any,
    ) -> None:
        """Initialize the BestAcquisition element.

        Parameters
        ----------
        itr_max : int | None, optional
            The maximum iteration to consider. If None, the current
            iteration is used.
        **scatter_kwargs : Any
            Keyword arguments passed to `matplotlib.axes.Axes.scatter`.
        """
        self.itr_max = itr_max or self.graphic.itr
        super().__init__(**scatter_kwargs)

    @property
    def scatter_data(self) -> NDArray:
        """The best acquisition data to be plotted."""
        graphic = self.graphic
        res = graphic.results
        x_best, y_best = res.get_best_acq(self.itr_max)
        best = np.hstack((x_best, y_best))
        return best[:, : graphic.NUM_AXIS]



import itertools
import warnings
from typing import Iterable

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.stats import qmc


class InitManager:
    def __init__(
        self,
        inittype: str,
        bounds: ArrayLike,
        initpts: int,
        seed: int | None = None,
        scramble: bool = False,
    ) -> None:
        """
        Creates initial points which can be queried with the get_x and get_all
        methods. Available types of initial points (parameter inittype) are:
            sobol
            random
            grid
        """
        bounds = np.atleast_2d(bounds)
        self.dim = bounds.shape[0]
        self.rng = np.random.default_rng(seed)
        if inittype.lower() == "sobol":
            self.X_init = self._sobol(bounds, initpts, scramble=scramble)
        elif inittype.lower() == "random":
            self.X_init = self._random(bounds, initpts)
        elif inittype.lower() == "grid":
            n_side = np.power(initpts, 1.0 / self.dim)
            if np.all(np.isclose(n_side, n_side.astype(int))):
                n_side = n_side.astype(int)
            else:
                n_side = np.round(n_side).astype(int)
                initpts = n_side**self.dim
                warnings.warn(
                    "Grid based initial point creation modifies"
                    + " initpts so that nth root of it is an integer"
                    + " , where n in the number of dimensions."
                )
            self.X_init = self._grid(bounds, n_side)
        else:
            raise TypeError(f"Unknown initial point type: {inittype}")

    def get_x(self, i: int) -> NDArray:
        """
        Returns the i:th initial point
        """
        return self.X_init[i, :]

    def get_all(self) -> NDArray:
        """
        Returns all generated initial points
        """
        return self.X_init

    def _sobol(
        self, bounds: NDArray, initpts: ArrayLike, scramble: bool = False
    ) -> NDArray:
        """
        Initial points with the quasi-random Sobol sequence
        """
        npts = np.max(initpts)
        sampler = qmc.Sobol(d=self.dim, scramble=scramble, seed=self.rng)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            sample = sampler.random(npts)
        pts = qmc.scale(sample, bounds[:, 0], bounds[:, 1])
        if not np.isscalar(initpts) and len(initpts) > 1:
            pts = self._extend(pts, initpts)
        return pts

    def _random(self, bounds: NDArray, initpts: int) -> NDArray:
        """
        Initial points randomly
        """
        npts = np.max(initpts)
        pts = self.rng.random((npts, self.dim))
        pts = pts * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]
        if not np.isscalar(initpts) and len(initpts) > 1:
            pts = self._extend(pts, initpts)
        return pts

    def _grid(self, bounds: NDArray, npts: ArrayLike) -> NDArray:
        """
        Initial points in a grid. Total number of points returned is npts^dim.
        """
        if not np.isscalar(npts) and len(npts) > 1:
            pts = []
            for task, task_npts in enumerate(npts):
                task_pts = self._make_grid(bounds, task_npts)
                pts.append(np.hstack((task_pts, np.full((len(task_pts), 1), task))))
            pts = np.vstack(pts)
        else:
            pts = self._make_grid(bounds, int(np.take(npts, 0)))
        return pts

    def _make_grid(self, bounds: NDArray, npts: int) -> NDArray:
        """
        Return grid with npts points across bounds in each dimension.
        """
        base = [np.linspace(*b, num=npts, endpoint=True) for b in bounds]
        return np.array(list(itertools.product(*base))).astype(float)

    def _extend(self, data: ArrayLike, npts: Iterable) -> NDArray:
        """
        Extend initial points with task index.
        """
        data = np.atleast_2d(data)
        data_extended = []
        for task, task_npts in enumerate(npts):
            task_data = np.hstack((data[:task_npts], np.full((task_npts, 1), task)))
            data_extended.append(task_data)
        data_extended = np.vstack(data_extended)
        return data_extended



from __future__ import annotations

from functools import reduce

import GPy
import numpy as np
from GPy.kern import Kern
from numpy.typing import ArrayLike, NDArray

import boss.bo.kernels.kernel_factory as kernel_factory
from boss.bo.models.single_task import SingleTaskModel
from boss.utils.arrays import shape_consistent_XY


class GradientModel(SingleTaskModel):
    """
    Functionality for creating, refitting and optimizing a GP model with
    gradient observations.

    The GradientModel utilizes the GPy MultioutputGP model class, which allows
    for multiple input and output channels. We can include observed gradient
    data in GPR by defining separate channels for partial derivatives, in
    addition to the main function value channel.

    The DiffKern kernel computes cross-covariances between channels.
    """

    def __init__(
        self,
        kernel: Kern,
        X: ArrayLike,
        Y_dY: ArrayLike,
        noise: float = 1e-12,
        ynorm: bool = False,
        noise_optim=False,
    ) -> None:
        """
        Initializes the GradientModel class.
        """
        self._dim = kernel.input_dim

        # input channels
        X_list = [X] * (self.dim + 1)

        # observations
        Y, dY = Y_dY[:, :1], Y_dY[:, 1:]
        # normalization
        self.use_norm = ynorm
        self.normmean = np.mean(Y)
        self.normsd = np.ptp(Y) if self.use_norm else 1
        Y_norm = (Y - self.normmean) / self.normsd
        # output channels
        Y_list = [Y_norm] + [dY[:, d, None] for d in range(self.dim)]

        # the kernel is accompanied with a DiffKern for each partial derivative.
        kernel_list = [kernel]
        kernel_list += [GPy.kern.DiffKern(kernel, d) for d in range(self.dim)]

        # noise is given to the likelihood.
        likelihood = GPy.likelihoods.Gaussian(variance=noise)
        likelihood_list = [likelihood] * (self.dim + 1)

        # initialize model
        self.model = GPy.models.MultioutputGP(
            X_list=X_list,
            Y_list=Y_list,
            kernel_list=kernel_list,
            likelihood_list=likelihood_list,
        )
        self.noise_optim = noise_optim
        if not noise_optim:
            self.model.likelihood.fix()

    @property
    def dim(self) -> int:
        return self._dim

    @property
    def kernel(self) -> Kern:
        return self.model.kern

    @property
    def X(self):
        X_multioutput = self.model.X[:, :-1]
        output_index = self.model.X[:, -1]

        return X_multioutput[np.where(output_index == 0)[0]]

    @property
    def Y(self):
        Y_multioutput = self.model.Y
        output_index = self.model.X[:, -1]

        Y_norm = Y_multioutput[np.where(output_index == 0)[0]]
        Y = Y_norm * self.normsd + self.normmean

        dY = np.empty((len(Y), self.dim), dtype=float)
        for d in range(self.dim):
            dY[:, d, None] = Y_multioutput[np.where(output_index == d + 1)[0]]

        return np.concatenate((Y, dY), axis=1)

    def add_data(self, X_new: ArrayLike, Y_dY_new: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset appending.
        """
        # construct new unnormalized dataset
        X_new, Y_dY_new = shape_consistent_XY(
            X_new, Y_dY_new, x_dim=self.dim, y_dim=self.dim + 1
        )
        X = np.vstack([self.X, X_new])
        Y_dY = np.vstack([self.Y, Y_dY_new])
        # update model
        self.redefine_data(X, Y_dY)

    def redefine_data(self, X: ArrayLike, Y_dY: ArrayLike) -> None:
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        Y, dY = Y_dY[:, :1], Y_dY[:, 1:]
        # update normalization
        self.normmean = np.mean(Y)
        if self.use_norm:
            self.normsd = np.ptp(Y)
        # update model
        Y_norm = (Y - self.normmean) / self.normsd
        X_list = [X] * (self.dim + 1)
        Y_list = [Y_norm] + [dY[:, d, None] for d in range(self.dim)]
        self.model.set_XY(X_list, Y_list)

    def get_best_xy(self) -> tuple[NDArray, float]:
        """
        Returns the lowest energy acquisition (x, y).
        """
        x_best = np.array(self.X[np.argmin(self.Y[:, 0])])
        y_best = np.min(self.Y[:, 0])
        return x_best, y_best

    def predict(
        self, X: ArrayLike, noise: bool = True, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise) and normalisation (norm).
        """
        m, v = self.model.predict([np.atleast_2d(X)], include_likelihood=noise)
        v = np.clip(v, 1e-12, np.inf)
        if norm:
            return m, v
        return m * self.normsd + self.normmean, v * (self.normsd**2)

    def predict_grads(
        self, X: ArrayLike, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance gradients with respect to
        input at point x, with or without normalisation (norm).
        """
        dmdx, dvdx = self.model.predictive_gradients([np.atleast_2d(X)])
        if norm:
            return dmdx[:, :, None], dvdx
        return (dmdx * self.normsd)[:, :, None], dvdx * (self.normsd**2)

    def estimate_num_local_minima(self, search_bounds: ArrayLike) -> int:
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        # For the ith dimension, the number of local minima along a slice
        # is approximately n(i) = boundlength(i)/(2*lengthscale(i)). Note
        # that periodic kernels operate on normalised distances: distance
        # between inputs that are period(i)/2 apart is 1. To get the total
        # number of minima for all of the search space, multiply together
        # n(i) over all i.
        numpts = 1

        # For the GradientModel, the self.model.kern is the
        # MultioutputDerivativeKern. If self.dim > 1, the Prod kernel which
        # contains the individual kernels is located by
        # self.model.kern.parts[0]. If self.dim == 1, the individual kernel is
        # located by self.model.kern.parts.
        if self.dim > 1:
            ks = self.model.kern.parts[0].parts
        else:
            ks = self.model.kern.parts
        for bounds, kern in zip(search_bounds, ks):
            if hasattr(kern, "period"):
                bound_distance = (bounds[1] - bounds[0]) / kern.period[0]
            else:
                bound_distance = (bounds[1] - bounds[0]) / 2
            numpts *= max(1, bound_distance / kern.lengthscale[0])
        return int(numpts)

    def get_all_params(self, include_fixed=True) -> dict[str, float | list]:
        # would have used super().get_all_params
        # but 1-dim gradient kernels have weird structure
        if self.kernel.input_dim == 1:
            kern = self.kernel.parts
        else:
            kern = self.kernel.parts[0]
            
        is_unfixed = lambda par: "fixed" not in par.constraints.properties()
        
        # Add kernel parameters, first get all parameter names,
        # these are formatted like 'kernel_name.param_name'
        param_names = kern.parameter_names()
        params = {}
        for name in param_names:
            # separate kernel_name from para_name and get the 
            # actual parameter object by repeatedly applying getattr
            obj_names = name.split('.')
            par = reduce(lambda x, y: getattr(x, y), obj_names, kern)
            if is_unfixed(par) or include_fixed:
                new_name = '.'.join(obj_names[-2:])
                new_name = super()._rename_param(new_name)
                params[new_name] = np.squeeze(par.values)

        if include_fixed or self.noise_optim:
            # The MultioutputGP model can contain multiple likelihoods
            # We only use one, and access the noise through model.likelihood[0]
            noise = self.model.likelihood[0]
            params["noise.variance"] = noise

        return params



import gpytorch
import torch
from gpytorch.constraints import Interval

from boss.bo.kernels.kernel_factory import (_resolve_thepriorpar,
                                            _resolve_thetainit)
from boss.settings import Settings


def select_kernel(settings: Settings) -> gpytorch.kernels:
    """Convenience function to initialize a kernel and apply hyperparameters."""
    kern = init_kernel(settings)
    apply_hyper_settings(kern, settings)
    return kern


def init_kernel(settings: Settings) -> gpytorch.kernels:
    """Creates a kernel according to the settings.

    The kernel will be a product kernel with parts specified by the
    kernel keyword, no priors or constraints will be set.
    """
    kernel_parts = [None] * (settings.dim)

    for i in range(settings.dim):
        ktype = settings["kernel"][i]
        if ktype == "stdp":
            kernel_parts[i] = gpytorch.kernels.PeriodicKernel(active_dims=[i])

        elif ktype == "rbf":
            kernel_parts[i] = gpytorch.kernels.RBFKernel(active_dims=[i])

        elif ktype == "mat32":
            kernel_parts[i] = gpytorch.kernels.MaternKernel(nu=1.5, active_dims=[i])

        elif ktype == "mat52":
            kernel_parts[i] = gpytorch.kernels.MaternKernel(nu=2.5, active_dims=[i])

        else:
            raise TypeError(f"Unknown kernel: {settings['kernel'][i]}")

    kern = kernel_parts[0]
    if len(kernel_parts) > 1:
        for i in range(1, len(kernel_parts)):
            kern = kern * kernel_parts[i]
    kernel = gpytorch.kernels.ScaleKernel(kern)

    return kernel


def apply_hyper_settings(kernel: gpytorch.kernels, settings: Settings) -> None:
    set_hyper_values(kernel, settings)
    set_hyper_priors(kernel, settings)
    set_hyper_constraints(kernel, settings)
    if settings.is_multi:
        raise TypeError(f"Multitask GpyTorch backend not supported yet!")


def set_hyper_values(kernel: gpytorch.kernels, settings: Settings) -> None:
    """
    Sets gyperparameter values
    """
    thetainit = _resolve_thetainit(settings)

    if isinstance(kernel.base_kernel, gpytorch.kernels.ProductKernel):
        kernel_parts = list(kernel.base_kernel.kernels)
    else:
        kernel_parts = list(kernel.sub_kernels())

    kernel._set_outputscale(
        torch.tensor(thetainit[0], dtype=torch.float, requires_grad=True)
    )

    for i, kpart in enumerate(kernel_parts):
        kpart[i]._set_lengthscale(
            torch.tensor(thetainit[i + 1], dtype=torch.float, requires_grad=True)
        )

        if settings["kernel"][i] == "stdp":
            kpart[i]._set_period_length(
                torch.tensor(
                    settings["periods"][i], dtype=torch.float, requires_grad=True
                )
            )


def set_hyper_priors(kernel: gpytorch.kernels, settings: Settings) -> None:
    """
    Sets hyperparameter priors on kernels.
    """
    # Instantiate prior of given type
    thetaprior = settings["thetaprior"]
    if thetaprior is None:
        return  # dont use a prior
    elif thetaprior == "gamma":
        set_gamma_prior(kernel, settings)
    else:
        raise TypeError(f"Prior type {thetaprior} not supported")


def set_gamma_prior(kernel: gpytorch.kernels, settings: Settings) -> None:
    """
    Sets a gamma prior on the kernel.
    """
    thetapriorpar = _resolve_thepriorpar(settings)
    # Now we can set the actual priors
    if isinstance(kernel.base_kernel, gpytorch.kernels.ProductKernel):
        kernel_parts = list(kernel.base_kernel.kernels)
    else:
        kernel_parts = list(kernel.sub_kernels())

    prior = gpytorch.priors.GammaPrior

    kernel.register_prior(
        "gamma_scale", prior(thetapriorpar[0, 0], thetapriorpar[0, 1]), "outputscale"
    )

    for i, kpart in enumerate(kernel_parts):
        kpart[i].register_prior(
            "gamma_length" + str(i),
            prior(
                thetapriorpar[i + 1, 0],
                thetapriorpar[i + 1, 1],
            ),
            "lengthscale",
        )


def set_hyper_constraints(kernel: gpytorch.kernels, settings: Settings) -> None:
    """
    Sets hyperparameter constraints on kernels.
    """
    if isinstance(kernel.base_kernel, gpytorch.kernels.ProductKernel):
        kernel_parts = list(kernel.base_kernel.kernels)
    else:
        kernel_parts = list(kernel.sub_kernels())

    # variance
    if settings["thetabounds"] is not None:  # by default not set
        kernel.outputscale_constraint = Interval(
            settings["thetabounds"][0][0], settings["thetabounds"][0][1]
        )

        for i, kpart in enumerate(kernel_parts):
            kpart.lengthscale_constraint = Interval(
                settings["thetabounds"][i + 1][0], settings["thetabounds"][i + 1][1]
            )

    # fix periods
    for i, kpart in enumerate(kernel_parts):
        if settings["kernel"][i] == "stdp":  # pbc
            kpart.period_length_constraint = Interval(
                settings["periods"][i], settings["periods"][i] + 1e-3
            )
            kpart.raw_period_length.requires_grad = False  # Fix period lengthscale



import numpy as np

from boss.bo.acq.acquisition import Acquisition


class LCB(Acquisition):
    """
    Lower Confidence Bound (LCB) acquisition function with constant
    exploration weight.


    Takes exploration weight as parameter.
    """

    def __init__(self, model=None, weight=2.0):
        """Takes one parameter to balance exploration-exploitation. Larger
        values favor exploration.

        Parameters
        ----------
        weight : float
            Parameter > 0 to boost exploration.
        """
        super().__init__(model=model)
        self.weight = weight

    def evaluate(self, x):
        """
        Compute the LCB acquisition function at x.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq : ndarray
          Acquisition function evaluated at 'x'. 2D array containing data with
          'float' type.
        """
        mean, var = self.model.predict(x, noise=False)
        f_acqu = mean - self.weight * np.sqrt(var)
        return f_acqu

    def evaluate_with_gradient(self, x):
        """
        Compute the LCB acquisition function and its gradient at x.
        Gradients are required to apply L-BFGS-B minimizer.

        Parameters
        ----------
        x : ndarray
          Input to the acquisition function. 2D array containing data with
          'float' type.

        Returns
        -------
        f_acq, df_acq : tuple of ndarray
          Acquisition function and it's gradient, evaluated at 'x'. Tuple of
          2D arrays, containing data with 'float' type.
        """
        mean, sd, dmean, dsd = self.model.predict_mean_sd_grads(x, noise=False)
        f_acqu = mean - self.weight * sd
        df_acqu = dmean - self.weight * dsd
        return f_acqu, df_acqu

    @property
    def has_gradient(self):
        """Property to inform whether acquisition function has a gradient.

        Returns
        -------
        bool
        """
        return True



import os
import platform
from enum import Enum
from pathlib import Path
from typing import Any, Literal

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import rcParams
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.typing import ArrayLike, NDArray


def set_legend(
    ax: Axes | None = None,
    unique: bool = False,
    shift: tuple[int, int] = (0, 0),
    box: bool = False,
    **kwargs: Any,
) -> None:
    """Configure and add a legend to a matplotlib axes.

    This function provides convenient legend positioning including special
    "outside" locations, optional shifting, and unique label filtering.

    Parameters
    ----------
    ax : Axes or None, optional
        The axes to add the legend to. If None, uses current axes.
    unique : bool, optional
        If True, filters out duplicate labels keeping only the first occurrence.
        Default is False.
    shift : tuple of int, optional
        A (dx, dy) shift to apply to the legend's bbox_to_anchor position.
        Default is (0, 0).
    box : bool, optional
        If True, draws a fancy box around the legend with transparency.
        If False, removes the legend frame. Default is False.
    **kwargs : Any
        Additional keyword arguments passed to ax.legend(). Special loc values
        include "outside right", "outside upper"/"outside top", and
        "outside lower"/"outside bottom" for positioning legends outside the plot area.
    """
    if box:
        rcParams["legend.fancybox"] = True
        rcParams["legend.frameon"] = True
        rcParams["legend.framealpha"] = 0.8
    else:
        rcParams["legend.frameon"] = False
        rcParams["legend.fancybox"] = False
        rcParams["legend.framealpha"] = None
        rcParams["legend.edgecolor"] = "inherit"

    if ax is None:
        ax = plt.gca()

    handles, labels = ax.get_legend_handles_labels()

    loc = kwargs.get("loc", None)
    if loc == "outside right":
        kwargs["loc"] = "center left"
        kwargs["bbox_to_anchor"] = np.array([1.0, 0.5])
        kwargs["bbox_transform"] = ax.transAxes
    elif loc in ["outside upper", "outside top"]:
        kwargs["loc"] = "lower center"
        kwargs["bbox_to_anchor"] = np.array([0.5, 1.0])
        kwargs["bbox_transform"] = ax.transAxes
        if not kwargs.get("ncol"):
            kwargs["ncol"] = len(labels)
    elif loc in ["outside lower", "outside bottom"]:
        kwargs["loc"] = "lower center"
        kwargs["bbox_to_anchor"] = np.array([0.5, -0.3])
        kwargs["bbox_transform"] = ax.transAxes
        if not kwargs.get("ncol"):
            kwargs["ncol"] = len(labels)

    bba = kwargs.get("bbox_to_anchor")
    if bba is not None:
        new_bba = np.asarray(bba) + shift
        kwargs["bbox_to_anchor"] = new_bba

    if unique:
        leg_uniq = {k: v for k, v in zip(labels, handles)}
        ax.legend(leg_uniq.values(), leg_uniq.keys(), **kwargs)
    else:
        ax.legend(**kwargs)


def set_grid(ax: Axes | None = None, seaborn_style: bool = True) -> None:
    """Add a grid to a matplotlib axes with optional seaborn styling.

    Parameters
    ----------
    ax : Axes or None, optional
        The axes to add the grid to. If None, uses current axes.
    seaborn_style : bool, optional
        If True, applies a seaborn-like grid style with light grey background
        and white grid lines. If False, uses matplotlib's default grid style.
        Default is True.
    """
    if ax is None:
        ax = plt.gca()

    if seaborn_style:
        ax.patch.set_facecolor("lightgrey")
        ax.patch.set_alpha(0.5)
        ax.grid(visible=True, color="white")
        ax.set_axisbelow(True)
    else:
        ax.grid(visible=True)


def set_conditional_logscale(
    ax: Axes,
    data: ArrayLike,
    which_axis: Literal["x", "y"] = "y",
    log_tol: float = 2.0,
) -> None:
    """Conditionally sets an axis to logscale.

    If the logarithmic range of data spans at least log_tol orders of
    magnitude the specified axis is set to logscale.
    """
    log_diff = np.log10(np.max(data)) - np.log10(np.min(data))
    if log_diff > log_tol:
        if which_axis.lower() == "y":
            ax.set_yscale("log")
        elif which_axis.lower() == "x":
            ax.set_xscale("log")


def scale_figsize(rows: float, cols: float) -> tuple[float, float]:
    """Determines an appropriate figsize for a matrix of subplots.

    The new size is obtained by scaling the default figsize by
    a requested number (can be a fraction) of rows and columns.

    Parameters
    ----------
    rows : float
        Number of subplot rows to scale by.
    cols : float
        Number of subplot columns to scale by.

    Returns
    -------
    tuple of float
        A (width, height) tuple for the scaled figure size.
    """
    size = matplotlib.rcParams["figure.figsize"]
    new_size = (cols * size[0], rows * size[1])
    return new_size



from __future__ import annotations

import copy

import gpytorch
import linear_operator.settings as linop_settings
import numpy as np
import torch
from numpy.typing import ArrayLike, NDArray

from boss.bo.models.model import Model


class SingleTaskModel(Model):
    """
    Functionality for creating, refitting and optimizing a GP model with Torch backend
    """

    def __init__(
        self,
        kernel: gpytorch.kernels.Kernel | None = None,
        X: ArrayLike | None = None,
        Y: ArrayLike | None = None,
        noise: float = 1e-12,
        ynorm: bool = False,
        precision: bool = True,
        noise_optim: bool = False,
    ):
        """
        Initializes the SingleTaskModel class.
        """
        if not torch.cuda.is_available():
            print("No GPU backend supported cuda device, defaulting to CPU")
        self.which_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.datatype = torch.float64  # Manual control over datatype
        torch.set_default_dtype(
            self.datatype
        )  # sets common torch dtype 32 faster but 64 to support scipy>=1.8
        self.ynorm = ynorm
        self.noise = noise
        self.precision = precision
        self.noise_optim = noise_optim

        if self.precision:
            # Increase default precision used by GpyTorch
            linop_settings._fast_covar_root_decomposition._default = False
            linop_settings._fast_log_prob._default = False
            linop_settings._fast_solves._default = False
            linop_settings.cholesky_max_tries._global_value = 6
            linop_settings.max_cholesky_size._global_value = 8192
            gpytorch.settings.max_eager_kernel_size._global_value = 8192

        self.reset(kernel, X, Y, self.noise, self.ynorm)

    def reset(self, kernel, X, Y, noise, ynorm):
        class GPyTorchRegression(
            gpytorch.models.ExactGP
        ):  # Inherit base model used for wrapping
            def __init__(self, train_x, train_y, likelihood, kernel):
                super().__init__(train_x, train_y, likelihood)
                self.mean = gpytorch.means.ConstantMean()
                self.covar_module = kernel

            def forward(self, x):  # Provide custom forward
                mean = self.mean(x)
                covar = self.covar_module(x)
                return gpytorch.distributions.MultivariateNormal(
                    mean, covar, validate_args=True
                )

        if X is not None and Y is not None:
            self.normmean = torch.mean(
                torch.tensor(Y), dtype=self.datatype
            )  # manually set dtype, inherits wrong dtype from numpy otherwise
            self.normsd = torch.tensor(np.ptp(Y)) if ynorm else 1.0
            X = torch.Tensor(X)
            Y = (torch.Tensor(Y).reshape(-1) - self.normmean) / self.normsd

        self.likelihood = gpytorch.likelihoods.GaussianLikelihood(
            noise_constraint=gpytorch.constraints.GreaterThan(1e-13)
        )
        self.model = GPyTorchRegression(
            X,
            Y,
            self.likelihood,
            kernel,
        )  # wrapper for model class
        self.likelihood.noise = noise
        # need to manually set... model and likelihood datatypes
        self.model = self.model.to(dtype=self.datatype)
        self.likelihood = self.likelihood.to(dtype=self.datatype)

        self.X_torch = X
        self.Y_torch = Y
        if self.X_torch is not None:
            self.X_torch = self.X_torch.to(self.which_device)
        if self.Y_torch is not None:
            self.Y_torch = self.Y_torch.to(self.which_device)
        self.likelihood.to(self.which_device)
        self.model.to(self.which_device)

        self.likelihood.noise_covar.raw_noise.requires_grad = self.noise_optim

        # Explicitly set this parameter to false so parameter methods can ignore it.
        self.model.mean.raw_constant.requires_grad = False

    @property
    def kernel(self):
        return self.model.covar_module

    @property
    def dim(self) -> int:
        return len(self._get_kernel_parts())

    @property
    def X(self) -> NDArray:
        if self.X_torch is not None:
            return np.atleast_2d(self.X_torch.cpu().numpy())
        else:
            return np.empty((0, self.dim), dtype=self.datatype)

    @property
    def Y(self) -> NDArray:
        if self.Y_torch is not None:
            Y = self.Y_torch * self.normsd + self.normmean
            return np.atleast_2d(Y.cpu().numpy().reshape(-1, 1))
        else:
            return np.empty((0, self.dim), dtype=self.datatype)

    def _get_kernel_parts(self):
        """
        Returns kernel(s) from the product kernel
        """
        if isinstance(self.kernel.base_kernel, gpytorch.kernels.ProductKernel):
            kernel_parts = list(self.kernel.base_kernel.kernels)
        else:
            kernel_parts = list(self.kernel.sub_kernels())
        return kernel_parts

    def predict(
        self, x: ArrayLike, noise: bool = True, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance at point x, with model variance (noise) as we return likelihood instead of model posterior.
        """
        x = np.atleast_2d(x)
        self.model.eval()
        self.likelihood.eval()
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            x_torch = torch.tensor(
                x, dtype=self.datatype, device=self.which_device
            )  # manually set dtype, otherwise inherits wrong type from numpy
            if noise:
                y_preds = self.likelihood(
                    self.model(x_torch)
                )  # equivalent to GPy inc likelihood = True, takes into acc epistemic uncertainty
            else:
                y_preds = self.model(
                    x_torch
                )  # only alleatoric uncertainty equivalent to GPy inc likelihood = False

            if norm:
                m = y_preds.mean
                v = y_preds.variance
            else:
                m = y_preds.mean * self.normsd + self.normmean
                v = y_preds.variance * (self.normsd**2)

        return m.cpu().numpy().reshape(-1, 1), v.cpu().numpy().reshape(-1, 1)

    def predict_grads(
        self, x: ArrayLike, norm: bool = False
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model prediction mean and variance gradients with respect to
        input at point x.
        """
        x = np.atleast_2d(x)
        x_torch = torch.tensor(
            x, dtype=self.datatype, requires_grad=True, device=self.which_device
        )
        self.model.eval()
        self.likelihood.eval()

        with gpytorch.settings.fast_pred_var():
            post = self.likelihood(self.model(x_torch))
            dmdx = torch.autograd.grad(post.mean.sum(), x_torch, retain_graph=True)[0]
            dvdx = torch.autograd.grad(post.variance.sum(), x_torch)[0]

        # Ensure shapes match Gpy return for EI acquisition function
        N, Q = x_torch.shape
        D = 1  # for single objective tasks
        dmdx = dmdx.view(N, Q, D)

        if norm:
            pass
        else:
            dmdx = dmdx * self.normsd
            dvdx = dvdx * (self.normsd**2)

        return (
            dmdx.cpu().numpy(),
            dvdx.cpu().numpy(),
        )

    def predict_mean_sd_grads(
        self, x: ArrayLike, noise: bool = True, norm: bool = True
    ) -> tuple[NDArray, NDArray, NDArray, NDArray]:
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise) and
        normalisation (norm).
        """
        m, v = self.predict(x, noise=noise, norm=norm)
        dmdx, dvdx = self.predict_grads(x, norm=norm)
        dmdx = dmdx[:, :, 0]
        dsdx = dvdx / (2 * np.sqrt(v))
        return m, np.sqrt(v), dmdx, dsdx

    def predict_mean_grad(
        self, x: ArrayLike, norm: bool = True
    ) -> tuple[NDArray, NDArray]:
        """
        Returns model mean and its gradient at point x, with or without
        normalisation (norm).
        """
        m, _ = self.predict(x, norm=norm)
        dmdx, _ = self.predict_grads(x, norm=norm)
        return m, dmdx

    def estimate_num_local_minima(self, search_bounds: ArrayLike) -> int:
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        numpts = 1
        ks = self._get_kernel_parts()

        for bounds, kern in zip(search_bounds, ks):
            if hasattr(kern, "period_length"):
                bound_distance = (bounds[1] - bounds[0]) / float(kern.period_length)
            else:
                bound_distance = (bounds[1] - bounds[0]) / 2

            numpts *= max(1, bound_distance / float(kern.lengthscale))

        return int(numpts)

    @staticmethod
    def _rename_param(name: str) -> str:
        """Renames parameters using BOSS conventions.

        Parameter names are changed from their interal gpytorch names to
        fit with BOSS conventions. These names are used by BOSS, e.g.,
        for writing to file.
        """
        substitutions = {
            "raw_": "",
            "outputscale": "variance",
            "base_kernel.kernels.": "kernel_",
            "base_kernel": "kernel",
            "period_length": "period",
        }
        new_name = name
        for text, sub in substitutions.items():
            new_name = new_name.replace(text, sub)

        return new_name

    def get_all_params(self, include_fixed: bool = True) -> dict[str, NDArray]:
        """
        Get all model parameter names and values.

        Parameter names returned this way are changed to conform
        with BOSS' nomenclature and do not in general correspond
        to the model-internal parameter names.

        Parameters
        ----------
        include_fixed : bool
            Whether to include fixed parameters, defaults to True.

        Returns
        -------
        dict[str, NDArray]
            Dict mapping parameter names to their corresponding values.
        """
        unfixed = self._get_param_names(include_fixed=True)
        params = {}
        for name, par in self.kernel.named_parameters():
            constraint = self.kernel.constraint_for_parameter_name(name)
            if par.requires_grad or include_fixed:
                new_name = self._rename_param(name)
                if constraint is not None:
                    actual_value = constraint.transform(par)
                else:
                    actual_value = par
                params[new_name] = np.squeeze(actual_value.data.cpu().numpy())

        # Add model noise last
        if self.noise_optim or include_fixed:
            noise = np.squeeze(self.likelihood.noise[0].cpu().detach().numpy())
            params["noise.variance"] = noise

        return params

    def get_unfixed_params(self) -> NDArray:
        """
        Returns the unfixed parameters of the model in an array.
        """
        unfixed = self._get_param_names()  # unfixed parameter names
        params = np.zeros(len(unfixed))
        for ind, param_name in enumerate(unfixed):
            current_attr = self.model
            for attr in param_name.split("."):
                current_attr = getattr(current_attr, attr)
            params[ind] = float(current_attr)
        return params

    def sample_unfixed_params(self, num_samples):
        """
        Sample unfixed model parameters.
        """
        raise NotImplementedError(
            "This method is not available in the current torch backend model."
        )

    def _get_param_names(self, include_fixed: bool = False) -> list[str]:
        """
        Get the model-internal names of all hyperparameters.
        """
        param_names = []
        for name, param in self.model.named_parameters():
            if param.requires_grad or include_fixed:
                param_names.append(name.replace("raw_", ""))
        return param_names

    def set_unfixed_params(
        self, params: NDArray
    ) -> None:  # setting hypers by object names!
        """
        Sets the unfixed parameters of the model to given values.
        """
        hypers = dict(zip(self._get_param_names(), torch.tensor(params)))
        self.model.initialize(**hypers)

    def _reinitialize_parameters(self, mean: bool = False) -> None:
        """
        Reinitialize model hyperparameters following gpy method of sampling from prior.
        """
        reinitialized_params = []
        for _, _, prior, _, _ in self.model.named_priors():
            if mean:
                reinitialized_params.append(prior.mean)
            else:
                reinitialized_params.append(prior.sample())
        self.set_unfixed_params(reinitialized_params)

    def optimize(
        self,
        restarts: int = 0,
        max_zero_grad_iter: int = 50,
        patience: int = 3,
        verbose: bool = False,
    ) -> None:
        """
        Updates the model hyperparameters by maximizing marginal likelihood with multiple restarts.
        """
        # Yuhao semi-heuristic values chosen for safe stability
        lr = 0.1  # this value is mulitplied to the optimized step size determined by the strong wolfe line search
        max_iter = 1000  # this value is the number of step size values that are tested for the strong wolfe conditions

        best_restart_loss = float("inf")
        best_state = None

        for restart in range(restarts):  # restarts helps prevent local minima trapping
            best_iter_loss = float("inf")  # reset best_iter_loss value after restart
            no_improve_count = 0  # reset counter after restart

            if (
                restart > 0
            ):  # first restart initialises from either thetainit (if new BOSS run) or most recently fitted hypers (if iterpts > 0)
                self._reinitialize_parameters(
                    mean=False
                )  # Other restarts draws hypers from their respective priors as initialisation points

            if verbose:
                print(f"restart {restart} with initialised model parameters:")
                print(self.get_all_params())

            try:
                self.model.train()
                self.likelihood.train()

                optimizer = torch.optim.LBFGS(
                    params=self.model.parameters(),
                    tolerance_grad=1e-7,
                    lr=lr,
                    max_iter=max_iter,
                    tolerance_change=1e-9,
                    history_size=2,
                    line_search_fn="strong_wolfe",
                )
                mll = gpytorch.mlls.ExactMarginalLogLikelihood(
                    self.likelihood, self.model
                )

                def closure():
                    self.model.train()
                    self.likelihood.train()
                    optimizer.zero_grad()
                    # Output prediction from model
                    output = self.model(self.X_torch)
                    # Calculate loss and backpropagate gradients
                    mll_loss = -mll(output, self.Y_torch)
                    loss = mll_loss
                    loss.backward()
                    return loss

                for count in range(max_zero_grad_iter):
                    optimizer.step(
                        closure
                    )  # optimizer carries out the strong wolfe line search at each optimizer iteration from which the step size at each iteration is determied

                    with torch.no_grad():
                        output = self.model(self.X_torch)
                        mll_loss = -mll(output, self.Y_torch)
                        iter_loss_value = mll_loss.item()

                    if verbose:
                        print(f"{iter_loss_value} of iteration {count}")

                    if iter_loss_value < best_iter_loss:
                        if verbose:
                            print("improvement")
                        best_iter_loss = iter_loss_value
                        no_improve_count = 0  # Reset patience increment counter if there is an improvement
                    else:
                        if verbose:
                            print("no improvement")
                        no_improve_count += (
                            1  # Add to patience increment counter if no improvement
                        )

                    # Break the loop if there is no improvement after 'patience' number of iterations
                    if no_improve_count >= patience:
                        if verbose:
                            print(
                                f"best iteration loss value: {best_iter_loss} found at iteration: {count} of restart: {restart} with patience of: {patience}"
                            )
                        break

                if verbose:
                    print(f" optimised after restart {restart} model parameters:")
                    print(self.get_all_params())

                # Evaluate the loss after optimization
                restart_loss_value = best_iter_loss

                # Update best loss and state
                if restart_loss_value < best_restart_loss:
                    best_restart_loss = restart_loss_value
                    if verbose:
                        print(
                            f"Current best restart loss value: {restart_loss_value} at restart {restart}"
                        )
                    best_state = {
                        "model_state": copy.deepcopy(self.model.state_dict()),
                        "likelihood_state": copy.deepcopy(self.likelihood.state_dict()),
                    }
            except Exception as e:
                if verbose:
                    print(f"Error during restart {restart}: {e}")
                continue

        # Load the best state after all restarts
        if best_state is not None:
            self.model.load_state_dict(best_state["model_state"])
            self.likelihood.load_state_dict(best_state["likelihood_state"])

            if verbose:
                print(f"best optimized params:{self.get_all_params()}")
        else:
            raise RuntimeError(
                "All model hyperparameter optimization restarts have failed"
            )

    def add_data(self, x_new, y_new):
        """
        Updates the model evidence (observations) dataset appending.
        """
        Y_unnorm = ((self.Y_torch * self.normsd) + self.normmean).cpu().numpy()
        added_X = np.vstack((self.X_torch.cpu().numpy(), x_new))
        added_Y = np.vstack((Y_unnorm.reshape(-1, 1), y_new))

        self.redefine_data(added_X, added_Y)

    def redefine_data(self, X, Y):
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        self.reset(self.kernel, X, Y, self.noise, self.ynorm)

    def get_best_xy(self):
        """
        Returns the lowest energy acquisitions (x, y).
        """
        x_best = self.X_torch[torch.argmin(self.Y_torch.reshape(-1, 1))]
        y_best = np.min(self.Y.reshape(-1, 1))
        return x_best.cpu().numpy(), y_best



"""High-level plotting interface for Bayesian optimization results.

This module provides the `Plotter` class, which serves as the main plotting
interface for post-processing BOSS optimization results. It is typically used
by the `PPMain` object (defined in `boss/pp/pp_main.py`) to generate various
diagnostic and analysis plots.

The module builds on top of the lower-level graphics infrastructure:
- `boss.pp.graphics`: Provides fundamental plotting primitives (`Curve`, `Contour`,
  `Surface`) that handle mesh-based visualization of functions over input spaces.
- `boss.pp.elements`: Defines composable plot elements (`Acquisitions`, `Minimum`,
  `NextAcquisition`) that can be overlaid on graphics.
- `boss.pp.mesh`: Handles coordinate grid generation and function evaluation.

The `Plotter` class encapsulates common plotting workflows and provides convenient
methods for visualizing:
- Acquisition history and convergence tracking
- Gaussian process models and uncertainty
- Acquisition functions
- True objective functions (when available)
- Hyperparameter evolution

Typical Usage
-------------
The `Plotter` is usually created and used via `PPMain`:

>>> from boss.pp.pp_main import PPMain
>>> pp = PPMain(results, pp_models=True, pp_acq_funcs=True)
>>> pp.run()

Direct usage is also supported:

>>> from boss.pp.plotter import Plotter
>>> plotter = Plotter.from_grid_spec(results)
>>> plotter.plot_model(itr=-1, show=True)
>>> plotter.plot_acquisitions(save_as="acquisitions.png")
"""

from __future__ import annotations

import itertools
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from numpy.typing import ArrayLike, NDArray

import boss.pp.style as style
import boss.pp.tools as tools
from boss.bo.results import BOResults
from boss.pp.elements import Acquisitions, Minimum, NextAcquisition
from boss.pp.graphics import Contour, Curve, MeshGraphic
from boss.pp.mesh import Mesh

Tableau20 = style.Tableau20


class Plotter:
    """High-level interface for creating plots from Bayesian optimization results.

    The `Plotter` class provides convenient methods for generating various diagnostic
    and analysis plots from a `BOResults` object. It uses `Mesh` objects to define
    coordinate grids and relies on the graphics primitives from `boss.pp.graphics`
    to create visualizations.

    Parameters
    ----------
    results : BOResults
        Bayesian optimization results object containing models, data, and settings.
    mesh : Mesh
        Mesh object defining the coordinate grid for plotting.
    style : str or Path or dict or None, optional
        Matplotlib style to use for plots. Can be a style name, path to style file,
        or dictionary of style parameters.

    Attributes
    ----------
    results : BOResults
        The BO results object.
    settings : Settings
        Settings object from the BO results.
    bounds : NDArray
        Optimization bounds from settings.
    colormap : str
        Default colormap for plots.
    mesh : Mesh
        The mesh object for coordinate grids.
    grid_spec : ArrayLike or None
        Grid specification from the mesh, if available.

    See Also
    --------
    boss.pp.pp_main.PPMain : Main post-processing interface
    boss.pp.graphics.MeshGraphic : Base class for mesh-based graphics
    boss.pp.mesh.Mesh : Coordinate grid generation
    """

    def __init__(
        self, results: BOResults, mesh: Mesh, style: str | Path | dict | None = None
    ) -> None:
        if style:
            plt.style.use(style)
        self.results = results
        self.settings = results.settings
        self.bounds = self.settings["bounds"]
        self.colormap = "viridis"
        self.mesh = mesh
        try:
            self.grid_spec = mesh.get_grid_spec()
        except RuntimeError:
            self.grid_spec = None

    @classmethod
    def from_grid_spec(
        cls,
        results: BOResults,
        grid_spec: ArrayLike | None = None,
        style: str | Path | dict | None = None,
    ):
        """Create a Plotter from a grid specification.

        This is a convenience constructor that creates a `Mesh` object from a grid
        specification before initializing the `Plotter`.

        Parameters
        ----------
        results : BOResults
            Bayesian optimization results object.
        grid_spec : ArrayLike or None, optional
            Grid specification for the mesh. If None, uses the grid spec from
            `results.settings["pp_model_slice"]`.
        style : str or Path or dict or None, optional
            Matplotlib style to use for plots.

        Returns
        -------
        Plotter
            Initialized plotter instance.
        """
        if grid_spec is None:
            grid_spec = results.settings["pp_model_slice"]
        mesh = Mesh.from_grid_spec(results.settings["bounds"], grid_spec)

        self = cls(results, mesh, style)
        return self

    def plot_acquisitions(
        self, save_as: str | Path | None = None, show: bool = True
    ) -> tuple[Figure, NDArray]:
        """Plot acquisition history over iterations.

        Creates a two-panel plot showing:
        - Top panel: Predicted minimum value with uncertainty bands and acquired y-values
        - Bottom panel: Acquisition locations and predicted minimum locations in input space

        Parameters
        ----------
        save_as : str or Path or None, optional
            Path to save the figure. If None, figure is not saved.
        show : bool, default=True
            Whether to display the figure.

        Returns
        -------
        fig : Figure
            The matplotlib figure object.
        ax_arr : NDArray
            Array of axes objects [ax1, ax2].
        """
        results = self.results
        iters_min = list(results["x_glmin"].keys())
        iters = results.batch_tracker.iteration_labels

        mu_glmin = results["mu_glmin"].to_array()
        nu_glmin = results["nu_glmin"].to_array()
        x_glmin = results["x_glmin"].to_array()

        colors = itertools.cycle(Tableau20)
        xticks = iters_min[:: max(1, round(len(iters_min) / 10))]
        fig, ax_arr = plt.subplots(
            2, 1, figsize=tools.scale_figsize(1.5, 1), sharex=True
        )
        ax1, ax2 = ax_arr

        ax1.fill_between(
            iters_min, mu_glmin + nu_glmin, mu_glmin, color=Tableau20.lightgrey
        )
        ax1.fill_between(
            iters_min, mu_glmin - nu_glmin, mu_glmin, color=Tableau20.lightgrey
        )
        ax1.plot(iters_min, mu_glmin + nu_glmin, Tableau20.grey)
        ax1.plot(iters_min, mu_glmin - nu_glmin, Tableau20.grey)
        ax1.plot(iters_min, mu_glmin, "black", label=r"$\mu(\hat{x})$")

        ax1.scatter(
            iters,
            results["Y"][:, 0],
            facecolors="none",
            edgecolors=next(colors),
            label=r"$acq_y$",
        )
        ax1.set_xlim(min(iters_min), max(iters_min))
        ax1.set_xticks(xticks)
        ax1.set_ylabel(r"$y$ with $\mu(\hat{x})$")

        for i in range(results.settings.dim):
            ax2.scatter(
                iters,
                results["X"][:, i],
                facecolors="none",
                edgecolors=next(colors),
                label=r"$acq_{x%i}$" % (i + 1),
            )

        for i in range(results.settings.dim):
            ax2.plot(
                iters_min,
                x_glmin[:, i],
                color=next(colors),
                label=r"$\hat{x}_%i$" % (i + 1),
            )

        ax2.set_xlim(iters_min[0], iters_min[-1])
        ax2.set_xticks(xticks)
        ax2.set_xlabel("Iteration")
        ax2.set_ylabel(r"$x_i$ with $\hat{x}_i$")
        fig.suptitle("Acquisitions")
        # --- unified save/show/close logic ---
        if save_as is not None:
            save_as = Path(save_as)
            fig.savefig(save_as, bbox_inches="tight")

        if show:
            fig.show()
        else:
            plt.close(fig)  # avoid auto-render in notebooks

        return fig, ax_arr

    def plot_convergence(
        self, save_as: str | Path | None = None, show: bool = True
    ) -> tuple[Figure, NDArray]:
        """Plot convergence diagnostics over iterations.

        Creates a two-panel plot showing convergence metrics:
        - Top panel: Change in predicted minimum location (Euclidean norm)
        - Bottom panel: Normalized change in predicted minimum value

        These metrics help assess whether the optimization has converged.

        Parameters
        ----------
        save_as : str or Path or None, optional
            Path to save the figure. If None, figure is not saved.
        show : bool, default=True
            Whether to display the figure.

        Returns
        -------
        fig : Figure
            The matplotlib figure object.
        ax_arr : NDArray
            Array of axes objects [ax1, ax2].
        """
        results = self.results
        iters_min = list(results["x_glmin"].keys())[1:]
        x_glmin = results["x_glmin"].to_array()
        mu_glmin = results["mu_glmin"].to_array()
        dmu_glmin = np.abs(np.diff(mu_glmin))
        dx_glmin = np.linalg.norm(np.diff(x_glmin, axis=0), axis=1)
        dy = np.zeros_like(dmu_glmin)
        for ind, itr in enumerate(iters_min):
            y_min, y_max = results.get_est_yrange(itr)
            dy[ind] = y_max - y_min

        fig, ax_arr = plt.subplots(
            2, 1, figsize=tools.scale_figsize(1.5, 1), sharex=True
        )
        ax1, ax2 = ax_arr
        xticks = iters_min[:: max(1, round(len(iters_min) / 10))]

        # dx_glmin
        ax1.plot(iters_min, dx_glmin, color=Tableau20.blue)
        ax1.set_ylabel(r"$\left|\Delta \hat{x}\right|$")
        ax1.set_xticks(xticks)
        tools.set_conditional_logscale(ax1, dx_glmin)

        # dmu_glmin
        ax2.plot(iters_min, dmu_glmin / dy, color=Tableau20.red)
        ax2.set_xlabel("Iteration")
        ax2.set_ylabel(r"$\left|\Delta \mu(\hat{x})\right|/\Delta y$")
        ax2.set_xticks(xticks)
        tools.set_conditional_logscale(ax2, dx_glmin)
        plt.suptitle("Convergence tracking")
        # ----- Unified save/show/close pattern -----
        if save_as is not None:
            save_as = Path(save_as)
            fig.savefig(save_as, bbox_inches="tight")

        if show:
            fig.show()
        else:
            plt.close(fig)  # prevents Jupyter inline auto-display

        return fig, ax_arr

    def plot_hyperparameters(
        self,
        save_as: str | Path | None = None,
        show: bool = True,
    ) -> tuple[Figure, NDArray]:
        """Plot Gaussian process hyperparameter evolution.

        Creates a two-panel plot showing how GP kernel hyperparameters change
        during optimization:
        - Top panel: Kernel variance over iterations
        - Bottom panel: Kernel lengthscales for each dimension over iterations

        Only plots hyperparameters that were not fixed during optimization.

        Parameters
        ----------
        save_as : str or Path or None, optional
            Path to save the figure. If None, figure is not saved.
        show : bool, default=True
            Whether to display the figure.

        Returns
        -------
        fig : Figure
            The matplotlib figure object.
        ax_arr : NDArray
            Array of axes objects [ax1, ax2].
        """
        results = self.results
        model_params = results["model_params"]
        iters = np.array(list(model_params.keys()))
        params = model_params.to_array(gaps=False)
        variances = params[:, 0]
        lengthscales = params[:, 1:]

        fig, ax_arr = plt.subplots(2, 1, figsize=tools.scale_figsize(2, 1), sharex=True)
        ax1, ax2 = ax_arr
        colors = itertools.cycle(Tableau20)
        markers = itertools.cycle(["s", "p", "*", "h", "^", "+", "x", "d", "v", "|"])

        # Variance
        ax1.plot(iters, variances, color=next(colors))
        ax1.set_ylabel("Variance")
        ax1.set_xticks(iters[:: max(1, round(len(iters) / 10))])
        tools.set_conditional_logscale(ax1, variances)

        # Lengthscales
        for i in range(results.settings.dim):
            ax2.plot(
                iters,
                lengthscales[:, i],
                color=next(colors),
                marker=next(markers),
                label=r"$\ell_ %i$" % (i + 1),
            )
        ax2.set_xlabel("Iteration")
        ax2.set_ylabel("Lengthscale")
        ax2.set_xticks(iters[:: max(1, round(len(iters) / 10))])
        tools.set_conditional_logscale(ax1, lengthscales.flatten())

        tools.set_legend(loc="best")
        plt.suptitle("Hyperparameter values")
        # --- NEW: consistent save/show/close logic ---
        if save_as is not None:
            save_as = Path(save_as)
            fig.savefig(save_as, bbox_inches="tight")

        if show:
            fig.show()
        else:
            plt.close(fig)   # prevents auto-display in notebooks

        return fig, ax_arr

    def plot_model(
        self,
        itr: int = -1,
        include_stdev: bool = True,
        save_as: str | Path | None = None,
        show: bool = True,
    ) -> list[MeshGraphic]:
        """Plot the Gaussian process model predictions.

        Creates visualizations of the GP model's mean predictions over the input
        space. For 1D problems, generates a curve plot; for higher dimensions,
        generates contour plots. Optionally includes uncertainty (standard deviation)
        visualizations.

        The plot includes overlays showing:
        - Acquisition points
        - Predicted global minimum
        - Next acquisition point (if available)

        Parameters
        ----------
        itr : int, default=-1
            Iteration number to plot. Negative indices count from the end.
        include_stdev : bool, default=True
            Whether to include uncertainty visualization. For 1D plots, shows
            uncertainty bands. For 2D plots, creates a separate contour plot.
        save_as : str or Path or None, optional
            Path to save the figure. If None, figure is not saved.
        show : bool, default=True
            Whether to display the figure.

        Returns
        -------
        list[MeshGraphic]
            List of graphic objects created (Curve or Contour instances).
            Contains 1-2 graphics depending on `include_stdev`.
        """

        model = self.results.reconstruct_model(itr=itr)
        mesh_dim = len(self.mesh.active_dims)
        bo_dim = len(self.mesh.bounds)

        # set defaults for fixed variables
        if var_defaults := self.settings["pp_var_defaults"]:
            dim_vals = {i: v for i, v in enumerate(var_defaults)}
            for i in self.mesh.active_dims:
                del dim_vals[i]
            self.mesh.fix_dims(dim_vals)
        else:
            self.mesh.fix_dim_preset(self.results, "min", itr=itr)

        Y, Y_var = self.mesh.evaluate_func(model.predict)
        Y_stdev = np.sqrt(Y_var)
        graphics = []

        if mesh_dim == 1:
            curve = Curve.from_mesh(
                results=self.results, target=Y, mesh=self.mesh, itr=itr
            )
            if bo_dim == 1:
                curve += Minimum(
                    s=30, marker="o", color=Tableau20.pink, label="Global min."
                )
                curve += Acquisitions(
                    s=30, marker="x", color=Tableau20.red, label="Acq."
                )

                X = self.results.get_next_acq(itr)
                curve.ax.axvline(x=X.item(), color=Tableau20.brown, label="Next acq.")
            else:
                # if we have fixed inputs to min the global min will
                # be on the 1d curve so we can plot it
                if curve.fixed_inputs == "min":
                    curve += Minimum(
                        s=30, marker="o", color=Tableau20.pink, label="Global min."
                    )

            if include_stdev:
                ax = curve.ax
                X = self.mesh.grid[0]
                ax.fill_between(X, Y + Y_stdev, Y, color=Tableau20.lightblue, alpha=0.7)
                ax.fill_between(X, Y - Y_stdev, Y, color=Tableau20.lightblue, alpha=0.7)

            curve.plot(
                grid="seaborn",
                ylabel=get_target_label("model_mean"),
                color=Tableau20.blue,
            )
            graphics.append(curve)

        else:
            cont = Contour.from_mesh(
                results=self.results,
                target=Y,
                mesh=self.mesh,
                itr=itr,
                fixed_inputs="min",
            )
            cont += Acquisitions(
                s=40, marker="x", color=Tableau20.red, label="Acquisitions"
            )
            cont += Minimum(s=40, marker="o", color=Tableau20.pink, label="Minimum")
            cont += NextAcquisition(
                s=40, marker="s", color=Tableau20.brown, label="Next Acquisition"
            )
            cont.plot()

            graphics.append(cont)
            if include_stdev:
                cont_stdev = Contour.from_mesh(
                    results=self.results,
                    target=Y_stdev,
                    mesh=self.mesh,
                    itr=itr,
                    fixed_inputs="min",
                )
                cont_stdev += Acquisitions(
                    s=40, marker="x", color=Tableau20.red, label="Acquisitions"
                )
                cont_stdev += Minimum(
                    s=40, marker="o", color=Tableau20.pink, label="Minimum"
                )
                cont_stdev += NextAcquisition(
                    s=40, marker="s", color=Tableau20.brown, label="Next Acquisition"
                )
                cont_stdev.plot()
                graphics.append(cont_stdev)
        # ---- handle all figures from all graphics ----
        figs: list[Figure] = []
        for g in graphics:
            if hasattr(g, "ax"):
                fig = g.ax.figure
                if fig not in figs:
                    figs.append(fig)

        # Saving
        if save_as is not None and figs:
            save_as = Path(save_as)
            # save first figure as requested
            figs[0].savefig(save_as, bbox_inches="tight")

            # if we have extra figures (e.g., stdev), save them with suffixes
            for idx, extra_fig in enumerate(figs[1:], start=1):
                # e.g. model_graph_0001_stdev.png
                extra_name = f"{save_as.stem}_stdev{idx}{save_as.suffix}"
                extra_path = save_as.with_name(extra_name)
                extra_fig.savefig(extra_path, bbox_inches="tight")

        # Showing / closing
        if show:
            for fig in figs:
                fig.show()
        else:
            for fig in figs:
                plt.close(fig)

        return graphics

    def plot_acq_func(
        self,
        itr=-1,
        save_as=None,
        show=True,
    ) -> MeshGraphic:
        """Plot the acquisition function.

        Visualizes the acquisition function over the input space at a specified
        iteration. For 1D problems, creates a curve plot; for higher dimensions,
        creates a contour plot. The plot includes overlays showing acquisition
        points, predicted minimum, and next acquisition point.

        Parameters
        ----------
        itr : int, default=-1
            Iteration number to plot. Negative indices count from the end.
        save_as : str or Path or None, optional
            Path to save the figure. If None, figure is not saved.
        show : bool, default=True
            Whether to display the figure.

        Returns
        -------
        MeshGraphic
            The graphic object created (Curve or Contour instance).
        """
        mesh_dim = len(self.mesh.active_dims)

        # set defaults for fixed variables
        if var_defaults := self.settings["pp_var_defaults"]:
            dim_vals = {i: v for i, v in enumerate(var_defaults)}
            for i in self.mesh.active_dims:
                del dim_vals[i]
            self.mesh.fix_dims(dim_vals)
        else:
            self.mesh.fix_dim_preset(self.results, "min", itr=itr)

        if mesh_dim == 1:
            graphic = Curve.from_mesh(
                results=self.results, target="acqfn", mesh=self.mesh, itr=itr
            )
            graphic.plot(
                grid="seaborn", ylabel=get_target_label("acqfn"), color=Tableau20.red
            )
        else:
            graphic = Contour.from_mesh(
                results=self.results,
                target="acqfn",
                mesh=self.mesh,
                itr=itr,
                fixed_inputs="min",
            )
            graphic += Acquisitions(
                s=40, marker="x", color=Tableau20.red, label="Acquisitions"
            )
            graphic += Minimum(s=40, marker="o", color=Tableau20.pink, label="Minimum")
            graphic += NextAcquisition(
                s=40, marker="s", color=Tableau20.brown, label="Next Acquisition"
            )
            graphic.plot()

        # get the figure from the graphic's axes
        fig = graphic.ax.figure if hasattr(graphic, "ax") else None

        if save_as is not None and fig is not None:
            save_as = Path(save_as)
            fig.savefig(save_as, bbox_inches="tight")

        if show and fig is not None:
            # fig.show() is a bit more explicit than global plt.show()
            fig.show()
        else:
            # close so Jupyter / inline backends don't auto-display
            if fig is not None:
                plt.close(fig)

        return graphic

    def plot_true_func(self, save_as: str | Path | None = None, show: bool = True):
        """Plot the true objective function.

        Visualizes the true objective function over the input space when it is
        available. For 1D problems, creates a curve plot; for higher dimensions,
        creates a contour plot. The plot includes overlays showing acquisition
        points and the predicted minimum.

        This is useful for comparing the GP model predictions against ground truth.

        Parameters
        ----------
        save_as : str or Path or None, optional
            Path to save the figure. If None, figure is not saved.
        show : bool, default=True
            Whether to display the figure.

        Returns
        -------
        MeshGraphic
            The graphic object created (Curve or Contour instance).
        """
        mesh_dim = len(self.mesh.active_dims)

        # set defaults for fixed variables
        if var_defaults := self.settings["pp_var_defaults"]:
            dim_vals = {i: v for i, v in enumerate(var_defaults)}
            for i in self.mesh.active_dims:
                del dim_vals[i]
            self.mesh.fix_dims(dim_vals)
        else:
            self.mesh.fix_dim_preset(self.results, "min", itr=-1)

        f_true = lambda X: self.settings.f.evaluate(X).Y
        Y = self.mesh.evaluate_func(f_true)

        if mesh_dim == 1:
            graphic = Curve.from_mesh(
                results=self.results,
                target=Y,
                mesh=self.mesh,
            )
            graphic.plot(
                grid="seaborn",
                ylabel="y",
                color=Tableau20.blue,
            )
        else:
            graphic = Contour.from_mesh(
                results=self.results,
                target=Y,
                mesh=self.mesh,
                fixed_inputs="min",
            )
            graphic += Acquisitions(
                s=40, marker="x", color=Tableau20.red, label="Acquisitions"
            )
            graphic += Minimum(
                s=40, marker="o", color=Tableau20.pink, label="Pred. Minimum"
            )
            graphic += NextAcquisition(
                s=40, marker="s", color=Tableau20.brown, label="Next Acquisition"
            )
            graphic.plot()

        # pick the figure from the first graphic (Curve or Contour)
        fig = None
        if graphic is not None and hasattr(graphic, "ax"):
            fig = graphic.ax.figure
        
        if save_as is not None and fig is not None:
            save_as = Path(save_as)
            fig.savefig(save_as, bbox_inches="tight")
        
        if show and fig is not None:
            fig.show()
        else:
            if fig is not None:
                plt.close(fig)
        
        return graphic


def get_target_label(
    target: Literal["model_mean", "model_var", "model_std", "acqfn"],
) -> str:
    """Get LaTeX label for a target function type.

    Parameters
    ----------
    target : {"model_mean", "model_var", "model_std", "acqfn"}
        Target function type.

    Returns
    -------
    str
        LaTeX-formatted label string for the target.
    """
    if target == "model_mean":
        label = r"$\mu$"
    elif target == "model_var":
        label = r"$\nu^2$"
    elif target == "model_std":
        label = r"$\nu$"
    elif target == "acqfn":
        label = r"$\alpha$"
    return label



__version__ = "1.13.3"
__url__ = "https://gitlab.com/cest-group/boss"
__description__ = "Bayesian optimization structure search"
__authors__ = [
    "Yuhao Zhang",
    "Ville Parkkinen",
    "Henri Paulamaki",
    "Arttu Tolvanen",
    "Ulpu Remes",
    "Nuutti Sten",
    "Emma Lehto",
    "Mikael Granit",
    "Manuel Kuchelmeister",
    "Tatu Linnala",
    "Tuomas Rossi",
    "Ransell Dsouza",
    "Matthias Stosiek",
    "Armi Tiihonen",
    "Joakim Loefgren",
    "Milica Todorovic",
]
__maintainer__ = "The BOSS developers team"
__maintainer_email__ = "milica.todorovic@utu.fi"
__license__ = "Apache License 2.0"



from functools import reduce
from typing import Any

import GPy
import numpy as np

from boss.settings import Settings
from boss.utils.distributions import gammaparams


def init_kernel(settings: Settings) -> GPy.kern.Kern:
    """Creates a kernel according to the settings.

    The kernel will be a product kernel with parts specified by the
    kernel keyword, no priors or constraints will be set.
    """
    kernel_parts = [None] * (settings.dim)

    for i in range(settings.dim):
        ktype = settings["kernel"][i]
        if ktype == "stdp":
            kernel_parts[i] = GPy.kern.StdPeriodic(
                input_dim=1,
                variance=1.0,
                period=1.0,
                lengthscale=1.0,
                ARD1=True,
                ARD2=True,
                active_dims=[i],
                name="kern",
            )
        elif ktype == "rbf":
            kernel_parts[i] = GPy.kern.RBF(
                input_dim=1,
                variance=1.0,
                lengthscale=1.0,
                ARD=True,
                active_dims=[i],
                name="kern",
            )
        elif ktype == "mat32":
            kernel_parts[i] = GPy.kern.Matern32(
                input_dim=1,
                variance=1.0,
                lengthscale=1.0,
                ARD=True,
                active_dims=[i],
                name="kern",
            )
        elif ktype == "mat52":
            kernel_parts[i] = GPy.kern.Matern52(
                input_dim=1,
                variance=1.0,
                lengthscale=1.0,
                ARD=True,
                active_dims=[i],
                name="kern",
            )
        else:
            raise TypeError(f"Unknown kernel: {settings['kernel'][i]}")

    # multiply all kernel parts together
    kernel = kernel_parts[0]
    if len(kernel_parts) > 1:
        for i in range(1, len(kernel_parts)):
            kernel = kernel * kernel_parts[i]

    # add coregionalisation
    if settings.is_multi:
        kernel = _add_coreg(kernel, settings)

    return kernel


def apply_hyper_settings(kernel: GPy.kern.Kern, settings: Settings) -> None:
    _set_hyper_values(kernel, settings)
    _set_hyper_priors(kernel, settings)
    _set_hyper_constraints(kernel, settings)
    if settings.is_multi:
        _apply_coreg_settings(kernel, settings)


def _set_hyper_constraints(kernel: GPy.kern.Kern, settings: Settings) -> None:
    """
    Sets hyperparameter constraints on kernels.
    """
    if isinstance(kernel, GPy.kern.Prod):
        kernel_parts = kernel.parts
    else:
        kernel_parts = [kernel]

    # variance
    if settings["thetabounds"] is not None:
        kernel_parts[0].variance.constrain_bounded(
            settings["thetabounds"][0][0],
            settings["thetabounds"][0][1],
            warning=False,
        )
        # lengthscale
        for i in range(settings.dim):
            kernel_parts[i].lengthscale.constrain_bounded(
                settings["thetabounds"][i + 1][0],
                settings["thetabounds"][i + 1][1],
                warning=False,
            )
    # fix periods
    for i in range(settings.dim):
        if settings["kernel"][i] == "stdp":  # pbc
            kernel_parts[i].period.constrain_fixed(
                settings["periods"][i], warning=False
            )

    # fix variances for all but the first kernel
    if settings.dim > 1:
        for i in range(1, settings.dim):
            kernel_parts[i].variance.constrain_fixed(1.0, warning=False)


def _resolve_thetainit(settings: Settings) -> np.ndarray:
    """
    Resolve the initial values of theta parameters based on the settings.

    Parameters:
    settings: Settings object containing configuration for initialization.

    Returns:
    list: A list of initial values for theta parameters.
    """
    thetainit = settings["thetainit"]
    if thetainit is None:
        if settings["ynorm"]:
            diff = 1.0  # when observations are range-normalized
        else:
            yrange = settings["yrange"]
            # An yrange estimate is typically either estimated from GP initial data
            # or provided directly by the user, but if we still don't have one at
            # this point we resort to guessing.
            if yrange is None:
                yrange = [-10.0, 10.0]
                settings["yrange"] = yrange

            diff = yrange[1] - yrange[0]
        thetainit = [0.5 * diff]  # sig
        for i in range(settings.dim):  # lengthscales
            if settings["kernel"][i] == "stdp":  # pbc
                thetainit.append(np.pi / 10)
            else:  # nonpbc
                thetainit.append(settings["periods"][i] / 20)
        thetainit = np.asarray(thetainit)
        settings["thetainit"] = (
            thetainit  # important to set new setting as coreg depends on it
        )
    return thetainit


def _resolve_thepriorpar(settings: Settings) -> np.ndarray:
    """
    Resolve the prior distribution values of the hyperparameters based on the settings.

    Parameters:
    settings: Settings object containing configuration for initialization.

    Returns:
    list: A nested list of prior distribution values for each hyperparameter.
    """
    thetapriorpar = settings["thetapriorpar"]
    if thetapriorpar is None:
        if settings["ynorm"]:
            diff = 1.0  # when observations are range normalised
        else:
            yrange = settings["yrange"]
            # An yrange estimate is typically either estimated from GP initial data
            # or provided directly by the user, but if we still don't have one at
            # this point we resort to guessing.
            if yrange is None:
                yrange = [-10.0, 10.0]
                settings["yrange"] = yrange
            diff = yrange[1] - yrange[0]

        # Ulpu's heuristic prior
        shape = 2.00
        rate = 2.0 / (diff / 2.0) ** 2

        # Original solution, to be tested further
        # shape, rate = Distributions.gammaparams(
        #    (diff/4)**2, (10*diff/4)**2, 0.5, 0.99)
        # shape = 1.0    # NORMALIZATION
        # rate = 1.5     # NORMALIZATION
        thetapriorpar = [[shape, rate]]

        for i in range(settings.dim):
            if settings["kernel"][i] == "stdp":  # pbc
                shape = 3.3678
                rate = 9.0204
            else:  # nonpbc
                shape, rate = gammaparams(
                    settings["periods"][i] / 20, settings["periods"][i] / 2
                )
            thetapriorpar.append([shape, rate])
        thetapriorpar = np.asarray(thetapriorpar)
        settings["thetapriorpar"] = (
            thetapriorpar  # important to set new setting as coreg depends on it
        )
    return thetapriorpar


def _set_hyper_values(kernel: GPy.kern.Kern, settings: Settings) -> None:
    """
    Sets hyperparameter values
    """
    thetainit = _resolve_thetainit(settings)

    if isinstance(kernel, GPy.kern.Prod):
        kernel_parts = kernel.parts
    else:
        kernel_parts = [kernel]

    # Set all hyperparameters for non-coreg kernels
    kernel_parts[0].variance = thetainit[0]
    lim = None
    if settings.is_multi:
        lim = -1  # skip coreg which is always last
    for i, kpart in enumerate(kernel_parts[:lim]):
        kpart.lengthscale = thetainit[i + 1]
        if settings["kernel"][i] == "stdp":
            kpart.period = settings["periods"][i]


def _set_hyper_priors(kernel: GPy.kern.Kern, settings: Settings) -> None:
    """
    Sets hyperparameter priors on kernels.
    """
    # Instantiate prior of given type
    thetaprior = settings["thetaprior"]
    if thetaprior is None:
        return  # dont use a prior
    elif thetaprior == "gamma":
        _set_gamma_prior(kernel, settings)
    else:
        raise TypeError(f"Prior type {thetaprior} not supported")


def _set_gamma_prior(kernel: GPy.kern.Kern, settings: Settings) -> None:
    """
    Sets a gamma prior on the kernel.
    """

    thetapriorpar = _resolve_thepriorpar(settings)

    # Now we can set the actual priors
    if isinstance(kernel, GPy.kern.Prod):
        kernel_parts = kernel.parts
    else:
        kernel_parts = [kernel]

    prior = GPy.priors.Gamma
    kernel_parts[0].variance.set_prior(
        prior(thetapriorpar[0, 0], thetapriorpar[0, 1]),
        warning=False,
    )
    for i in range(settings.dim):
        kernel_parts[i].lengthscale.set_prior(
            prior(
                thetapriorpar[i + 1, 0],
                thetapriorpar[i + 1, 1],
            ),
            warning=False,
        )


def _add_coreg(kernel: GPy.kern.Kern, settings: Settings) -> GPy.kern.Kern:
    """
    Extends a basic kernel with coregionalisation.
    """
    # note: input is one kernel rather than kernel list. this is on
    # purpose as the model class is not equipped to handle other cases
    # atm
    kernel_list = [kernel]
    output_dim = settings["num_tasks"]
    kernel_multi = GPy.util.multioutput.LCM(
        settings.dim, output_dim, kernel_list, settings["W_rank"]
    )
    return kernel_multi


def _apply_coreg_settings(kernel_multi: GPy.kern.Kern, settings: Settings) -> None:
    """
    Sets coregionalisation values and priors.
    """
    # fix base kernel variance to 1
    kernel_multi.parameters[0].variance.unset_priors()
    kernel_multi.parameters[0].variance.constrain_fixed(1.0)

    # set initial values
    kappa_init = settings["kappa_init"]
    if kappa_init is None:
        kappa_init = settings["thetainit"][0] * np.ones(settings["num_tasks"])
    settings["kappa_init"] = kappa_init

    new_shape = (settings["num_tasks"], settings["W_rank"])
    kernel_multi.parameters[-1].W = settings["W_init"].reshape(new_shape)
    kernel_multi.parameters[-1].kappa = settings["kappa_init"]

    # set priors
    if settings["W_prior"] is not None:
        param = kernel_multi.parameters[-1].W
        pname = settings["W_prior"]
        ppars = settings["W_priorpar"]
        if pname == "fixed_value":
            param.constrain_fixed(ppars[0])
        elif pname == "gaussian":
            param.set_prior(GPy.priors.Gaussian(ppars[0], ppars[1]), warning=False)
        else:
            raise TypeError("Unknown W_prior: {}".format(pname))
    if settings["kappa_prior"] is not None:
        param = kernel_multi.parameters[-1].kappa
        pname = settings["kappa_prior"]
        ppars = settings["kappa_priorpar"]
        if pname == "fixed_value":
            param.constrain_fixed(ppars[0])
        elif pname == "gamma":
            param.set_prior(GPy.priors.Gamma(ppars[0], ppars[1]), warning=False)
        else:
            raise TypeError("Unknown kappa_prior: {}".format(pname))
